import json
import re
import unittest
from unittest import mock

from pyramid import testing

from kinto import main as kinto_main
from kinto.core.testing import get_user_headers, skip_if_no_statsd

from .. import support


DATETIME_REGEX = r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}\+\d{2}:\d{2}$"


class PluginSetup(unittest.TestCase):
    @skip_if_no_statsd
    def test_a_statsd_timer_is_used_for_history_if_configured(self):
        settings = {
            "statsd_url": "udp://127.0.0.1:8125",
            "includes": "kinto.plugins.history",
        }
        config = testing.setUp(settings=settings)
        with mock.patch("kinto.core.statsd.Client.timer") as mocked:
            kinto_main(None, config=config)
            mocked.assert_called_with("plugins.history")


class HistoryWebTest(support.BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["includes"] = "kinto.plugins.history"
        return settings


class HelloViewTest(HistoryWebTest):
    def test_history_capability_if_enabled(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("history", capabilities)


class HistoryViewTest(HistoryWebTest):
    def setUp(self):
        self.bucket_uri = "/buckets/test"
        self.app.put(self.bucket_uri, headers=self.headers)

        self.collection_uri = self.bucket_uri + "/collections/col"
        resp = self.app.put(self.collection_uri, headers=self.headers)
        self.collection = resp.json["data"]

        self.group_uri = self.bucket_uri + "/groups/grp"
        body = {"data": {"members": ["elle"]}}
        resp = self.app.put_json(self.group_uri, body, headers=self.headers)
        self.group = resp.json["data"]

        self.record_uri = "/buckets/test/collections/col/records/rec"
        body = {"data": {"foo": 42}}
        resp = self.app.put_json(self.record_uri, body, headers=self.headers)
        self.record = resp.json["data"]

        self.history_uri = "/buckets/test/history"

    def test_only_get_and_delete_on_collection_are_allowed(self):
        self.app.put(self.history_uri, headers=self.headers, status=405)
        self.app.patch(self.history_uri, headers=self.headers, status=405)

    def test_only_collection_endpoint_is_available(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        url = "{}/{}".format(self.bucket_uri, entry["id"])
        self.app.get(url, headers=self.headers, status=404)
        self.app.put(url, headers=self.headers, status=404)
        self.app.patch(url, headers=self.headers, status=404)
        self.app.delete(url, headers=self.headers, status=404)

    def test_tracks_user_and_date(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][-1]
        assert entry["user_id"] == self.principal
        assert re.match(DATETIME_REGEX, entry["date"])

    #
    # Bucket
    #

    def test_history_contains_bucket_creation(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][-1]
        assert entry["resource_name"] == "bucket"
        assert entry["bucket_id"] == "test"
        assert entry["action"] == "create"
        assert entry["uri"] == "/buckets/test"

    def test_history_supports_creation_via_plural_endpoint(self):
        resp = self.app.post_json("/buckets", {"data": {"id": "posted"}}, headers=self.headers)
        resp = self.app.get("/buckets/posted/history", headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["resource_name"] == "bucket"
        assert entry["bucket_id"] == "posted"
        assert entry["action"] == "create"
        assert entry["uri"] == "/buckets/posted"

    def test_tracks_bucket_attributes_update(self):
        body = {"data": {"foo": "baz"}}
        self.app.patch_json(self.bucket_uri, body, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "update"
        assert entry["target"]["data"]["foo"] == "baz"

    def test_tracks_bucket_permissions_update(self):
        body = {"permissions": {"read": ["admins"]}}
        self.app.patch_json(self.bucket_uri, body, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "update"
        assert entry["target"]["permissions"]["read"] == ["admins"]

    def test_bucket_delete_destroys_its_history_entries(self):
        self.app.delete(self.bucket_uri, headers=self.headers)
        storage = self.app.app.registry.storage
        stored_in_backend = storage.list_all(parent_id="/buckets/test", resource_name="history")
        assert len(stored_in_backend) == 0

    def test_delete_all_buckets_destroys_history_entries(self):
        self.app.put_json("/buckets/1", {"data": {"a": 1}}, headers=self.headers)

        self.app.delete("/buckets?a=1", headers=self.headers)

        # Entries about deleted bucket are gone.
        storage = self.app.app.registry.storage
        stored_in_backend = storage.list_all(parent_id="/buckets/1", resource_name="history")
        assert len(stored_in_backend) == 0

        # Entries of other buckets are still here.
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][-1]
        assert entry["bucket_id"] == "test"
        assert entry["action"] == "create"

    #
    # Collection
    #

    def test_tracks_collection_creation(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][2]
        cid = self.collection["id"]
        assert entry["resource_name"] == "collection"
        assert "bucket_id" not in entry
        assert entry["collection_id"] == cid
        assert entry["action"] == "create"
        assert entry["uri"] == "/buckets/test/collections/{}".format(cid)

    def test_tracks_collection_attributes_update(self):
        body = {"data": {"foo": "baz"}}
        self.app.patch_json(self.collection_uri, body, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "update"
        assert entry["target"]["data"]["foo"] == "baz"

    def test_tracks_collection_permissions_update(self):
        body = {"permissions": {"read": ["admins"]}}
        self.app.patch_json(self.collection_uri, body, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "update"
        assert entry["target"]["permissions"]["read"] == ["admins"]

    def test_tracks_collection_delete(self):
        self.app.delete(self.collection_uri, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "delete"
        assert entry["target"]["data"]["deleted"] is True

    def test_tracks_multiple_collections_delete(self):
        self.app.put(self.bucket_uri + "/collections/col2", headers=self.headers)

        self.app.delete(self.bucket_uri + "/collections", headers=self.headers)

        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "delete"
        assert entry["target"]["data"]["id"] in (self.collection["id"], "col2")

    #
    # Group
    #

    def test_tracks_group_creation(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][1]
        assert entry["resource_name"] == "group"
        assert "bucket_id" not in entry
        assert entry["group_id"] == self.group["id"]
        assert entry["action"] == "create"
        assert entry["uri"] == "/buckets/test/groups/{}".format(self.group["id"])

    def test_tracks_group_attributes_update(self):
        body = {"data": {"foo": "baz", "members": ["lui"]}}
        self.app.patch_json(self.group_uri, body, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "update"
        assert entry["target"]["data"]["foo"] == "baz"
        assert entry["target"]["data"]["members"] == ["lui"]

    def test_tracks_group_permissions_update(self):
        body = {"permissions": {"read": ["admins"]}}
        self.app.patch_json(self.group_uri, body, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "update"
        assert entry["target"]["permissions"]["read"] == ["admins"]

    def test_tracks_group_delete(self):
        self.app.delete(self.group_uri, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "delete"
        assert entry["target"]["data"]["deleted"] is True

    def test_tracks_multiple_groups_delete(self):
        self.app.put_json(
            self.bucket_uri + "/groups/g2", {"data": {"members": ["her"]}}, headers=self.headers
        )

        self.app.delete(self.bucket_uri + "/groups", headers=self.headers)

        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "delete"
        assert entry["target"]["data"]["id"] in (self.group["id"], "g2")

    #
    # Record
    #

    def test_tracks_record_creation(self):
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        cid = self.collection["id"]
        rid = self.record["id"]
        assert entry["resource_name"] == "record"
        assert "bucket_id" not in entry
        assert entry["collection_id"] == cid
        assert entry["record_id"] == rid
        assert entry["action"] == "create"
        assert entry["uri"] == "/buckets/test/collections/{}/records/{}".format(cid, rid)
        assert entry["target"]["data"]["foo"] == 42
        assert entry["target"]["permissions"]["write"][0].startswith("basicauth:")

    def test_tracks_record_attributes_update(self):
        resp = self.app.patch_json(self.record_uri, {"data": {"foo": "baz"}}, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "update"
        assert entry["target"]["data"]["foo"] == "baz"

    def test_tracks_record_permissions_update(self):
        body = {"permissions": {"read": ["admins"]}}
        resp = self.app.patch_json(self.record_uri, body, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "update"
        assert entry["target"]["permissions"]["read"] == ["admins"]

    def test_tracks_record_delete(self):
        resp = self.app.delete(self.record_uri, headers=self.headers)
        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "delete"
        assert entry["target"]["data"]["deleted"] is True

    def test_tracks_multiple_records_delete(self):
        records_uri = self.collection_uri + "/records"
        body = {"data": {"foo": 43}}
        resp = self.app.post_json(records_uri, body, headers=self.headers)
        rid = resp.json["data"]["id"]

        self.app.delete(records_uri, headers=self.headers)

        resp = self.app.get(self.history_uri, headers=self.headers)
        entry = resp.json["data"][0]
        assert entry["action"] == "delete"
        assert entry["target"]["data"]["id"] in (self.record["id"], rid)

    def test_does_not_track_records_during_massive_deletion(self):
        body = {"data": {"pim": "pam"}}
        records_uri = self.collection_uri + "/records"
        self.app.post_json(records_uri, body, headers=self.headers)

        self.app.delete(self.collection_uri, headers=self.headers)

        resp = self.app.get(self.history_uri, headers=self.headers)
        deletion_entries = [e for e in resp.json["data"] if e["action"] == "delete"]
        assert len(deletion_entries) == 1


class HistoryDeletionTest(HistoryWebTest):
    def setUp(self):
        self.app.put("/buckets/bid", headers=self.headers)
        self.app.put("/buckets/bid/collections/cid", headers=self.headers)
        body = {"data": {"foo": 42}}
        self.app.put_json("/buckets/bid/collections/cid/records/rid", body, headers=self.headers)

    def test_full_deletion(self):
        self.app.delete("/buckets/bid/history", headers=self.headers)
        resp = self.app.get("/buckets/bid/history", headers=self.headers)
        assert len(resp.json["data"]) == 0

    def test_partial_deletion(self):
        resp = self.app.get("/buckets/bid/history", headers=self.headers)
        before = int(json.loads(resp.headers["ETag"]))
        self.app.put("/buckets/bid/collections/cid2", headers=self.headers)

        # Delete everything before the last entry (exclusive)
        self.app.delete("/buckets/bid/history?_before={}".format(before), headers=self.headers)

        resp = self.app.get("/buckets/bid/history", headers=self.headers)
        assert len(resp.json["data"]) == 2  # record + new collection


class FilteringTest(HistoryWebTest):
    def setUp(self):
        self.app.put("/buckets/bid", headers=self.headers)
        self.app.put("/buckets/0", headers=self.headers)
        self.app.put("/buckets/bid/collections/cid", headers=self.headers)
        self.app.put("/buckets/0/collections/1", headers=self.headers)
        body = {"data": {"foo": 42}}
        self.app.put_json("/buckets/bid/collections/cid/records/rid", body, headers=self.headers)
        body = {"data": {"foo": 0}}
        self.app.put_json("/buckets/0/collections/1/records/2", body, headers=self.headers)
        body = {"data": {"foo": "bar"}}
        self.app.patch_json("/buckets/bid/collections/cid/records/rid", body, headers=self.headers)
        self.app.delete("/buckets/bid/collections/cid/records/rid", headers=self.headers)

    def test_filter_by_unknown_field_is_not_allowed(self):
        self.app.get("/buckets/bid/history?movie=bourne", headers=self.headers, status=400)

    def test_filter_by_action(self):
        resp = self.app.get("/buckets/bid/history?action=delete", headers=self.headers)
        assert len(resp.json["data"]) == 1

    def test_filter_by_resource(self):
        resp = self.app.get("/buckets/bid/history?resource_name=bucket", headers=self.headers)
        assert len(resp.json["data"]) == 1

    def test_filter_by_uri(self):
        uri = "/buckets/bid/collections/cid/records/rid"
        resp = self.app.get("/buckets/bid/history?uri={}".format(uri), headers=self.headers)
        assert len(resp.json["data"]) == 3  # create / update / delete

    def test_allows_diff_between_two_versions_of_a_record(self):
        uri = "/buckets/bid/collections/cid/records/rid"
        querystring = "?uri={}&_limit=2&_sort=last_modified".format(uri)
        resp = self.app.get("/buckets/bid/history{}".format(querystring), headers=self.headers)
        entries = resp.json["data"]
        version1 = entries[0]["target"]["data"]
        version2 = entries[1]["target"]["data"]
        diff = [
            (k, version1[k], version2[k])
            for k in version1.keys()
            if k != "last_modified" and version2[k] != version1[k]
        ]
        assert diff == [("foo", 42, "bar")]

    def test_filter_by_bucket(self):
        uri = "/buckets/bid/history?bucket_id=bid"
        resp = self.app.get(uri, headers=self.headers)
        # This is equivalent to filtering by resource_name=bucket,
        # since only entries for bucket have ``bucket_id`` attribute.
        assert len(resp.json["data"]) == 1

    def test_filter_by_collection(self):
        uri = "/buckets/bid/history?collection_id=cid"
        resp = self.app.get(uri, headers=self.headers)
        assert len(resp.json["data"]) == 4

    def test_filter_by_numeric_bucket(self):
        uri = "/buckets/0/history?bucket_id=0"
        resp = self.app.get(uri, headers=self.headers)
        assert len(resp.json["data"]) == 1

    def test_filter_by_numeric_collection(self):
        uri = "/buckets/0/history?collection_id=1"
        resp = self.app.get(uri, headers=self.headers)
        assert len(resp.json["data"]) == 2

    def test_filter_by_numeric_record(self):
        uri = "/buckets/0/history?record_id=2"
        resp = self.app.get(uri, headers=self.headers)
        assert len(resp.json["data"]) == 1

    def test_filter_by_target_fields(self):
        uri = "/buckets/bid/history?target.data.id=rid"
        resp = self.app.get(uri, headers=self.headers)
        assert len(resp.json["data"]) == 3  # create, update, delete

    def test_limit_results(self):
        resp = self.app.get("/buckets/bid/history?_limit=2", headers=self.headers)
        assert len(resp.json["data"]) == 2
        assert "Next-Page" in resp.headers

    def test_filter_returned_fields(self):
        resp = self.app.get("/buckets/bid/history?_fields=uri,action", headers=self.headers)
        assert sorted(resp.json["data"][0].keys()) == ["action", "id", "last_modified", "uri"]

    def test_sort_by_date(self):
        resp = self.app.get("/buckets/bid/history?_sort=date", headers=self.headers)
        entries = resp.json["data"]
        assert entries[0]["date"] < entries[-1]["date"]


class BulkTest(HistoryWebTest):
    def setUp(self):
        body = {
            "defaults": {"method": "POST", "path": "/buckets/bid/collections/cid/records"},
            "requests": [
                {"path": "/buckets/bid", "method": "PUT"},
                {"path": "/buckets/bid/collections", "body": {"data": {"id": "cid"}}},
                {"body": {"data": {"id": "a", "attr": 1}}},
                {"body": {"data": {"id": "b", "attr": 2}}},
                {"body": {"data": {"id": "c", "attr": 3}}},
            ],
        }
        self.app.post_json("/batch", body, headers=self.headers)

    def test_post_on_collection(self):
        resp = self.app.get("/buckets/bid/history", headers=self.headers)
        entries = resp.json["data"]
        assert len(entries) == 5
        assert entries[0]["uri"] == "/buckets/bid/collections/cid/records/c"
        assert entries[-2]["uri"] == "/buckets/bid/collections/cid"

    def test_delete_on_collection(self):
        body = {
            "defaults": {"method": "DELETE"},
            "requests": [
                {"path": "/buckets/bid/collections/cid/records/a"},
                {"path": "/buckets/bid/collections/cid/records/b"},
                {"path": "/buckets/bid/collections/cid/records/c"},
            ],
        }
        self.app.post_json("/batch", body, headers=self.headers)
        resp = self.app.get("/buckets/bid/history", headers=self.headers)
        entries = resp.json["data"]
        assert entries[0]["uri"] == "/buckets/bid/collections/cid/records/c"
        assert entries[1]["uri"] == "/buckets/bid/collections/cid/records/b"
        assert entries[2]["uri"] == "/buckets/bid/collections/cid/records/a"

    def test_multiple_patch(self):
        # Kinto/kinto#942
        requests = [
            {
                "method": "PATCH",
                "path": "/buckets/bid/collections/cid/records/{}".format(label),
                "body": {"data": {"label": label}},
            }
            for label in ("a", "b", "c")
        ]
        self.app.post_json("/batch", {"requests": requests}, headers=self.headers)
        resp = self.app.get("/buckets/bid/history", headers=self.headers)
        entries = resp.json["data"]
        for entry in entries:
            if entry["resource_name"] != "record":
                continue
            assert entry["record_id"] == entry["target"]["data"]["id"]


class DefaultBucketTest(HistoryWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["includes"] = "kinto.plugins.default_bucket " "kinto.plugins.history"
        return settings

    def setUp(self):
        resp = self.app.get("/", headers=self.headers)
        self.bucket_id = resp.json["user"]["bucket"]
        self.history_uri = "/buckets/{}/history".format(self.bucket_id)

    def test_history_can_be_accessed_via_default_alias(self):
        self.app.get("/buckets/default/collections/blah", headers=self.headers)
        resp = self.app.get("/buckets/default/history", headers=self.headers)
        assert len(resp.json["data"]) == 2

    def test_implicit_creations_are_listed(self):
        body = {"data": {"foo": 42}}
        resp = self.app.post_json(
            "/buckets/default/collections/blah/records", body, headers=self.headers
        )
        record = resp.json["data"]

        resp = self.app.get(self.history_uri, headers=self.headers)
        entries = resp.json["data"]
        assert len(entries) == 3

        bucket_uri = "/buckets/{}".format(self.bucket_id)
        assert entries[2]["resource_name"] == "bucket"
        assert entries[2]["bucket_id"] == self.bucket_id
        assert entries[2]["uri"] == bucket_uri
        assert entries[2]["target"]["permissions"]["write"][0] == self.principal

        collection_uri = bucket_uri + "/collections/blah"
        assert entries[1]["resource_name"] == "collection"
        assert "bucket_id" not in entries[1]
        assert entries[1]["collection_id"] == "blah"
        assert entries[1]["uri"] == collection_uri
        assert entries[1]["target"]["permissions"]["write"][0] == self.principal

        record_uri = collection_uri + "/records/{}".format(record["id"])
        assert entries[0]["resource_name"] == "record"
        assert "bucket_id" not in entries[1]
        assert entries[0]["collection_id"] == "blah"
        assert entries[0]["record_id"] == record["id"]
        assert entries[0]["uri"] == record_uri
        assert entries[0]["target"]["data"]["foo"] == 42
        assert entries[0]["target"]["permissions"]["write"][0] == self.principal


class PermissionsTest(HistoryWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["experimental_permissions_endpoint"] = "true"
        return settings

    def setUp(self):
        self.alice_headers = get_user_headers("alice")
        self.bob_headers = get_user_headers("bob")
        self.julia_headers = get_user_headers("julia")

        self.alice_principal = (
            "basicauth:d5b0026601f1b251974e09548d44155e16" "812e3c64ff7ae053fe3542e2ca1570"
        )
        self.bob_principal = (
            "basicauth:c031ced27503f788b102ca54269a062ec737" "94bb075154c74a0d4311e74ca8b6"
        )
        self.julia_principal = (
            "basicauth:d8bab8d9fe0510fcaf9b5ad5942c027fc" "2fdf80b6dc59cc3c48d12a2fcb18f1c"
        )

        bucket = {"permissions": {"read": [self.alice_principal]}}
        collection = {"permissions": {"read": [self.julia_principal]}}
        record = {"permissions": {"write": [self.bob_principal, self.alice_principal]}}
        self.app.put("/buckets/author-only", headers=self.headers)
        self.app.put_json("/buckets/test", bucket, headers=self.headers)
        self.app.put_json(
            "/buckets/test/groups/admins", {"data": {"members": []}}, headers=self.headers
        )
        self.app.put_json(
            "/buckets/test/collections/alice-julia", collection, headers=self.headers
        )
        self.app.put_json("/buckets/test/collections/author-only", headers=self.headers)
        self.app.post_json(
            "/buckets/test/collections/alice-julia/records", record, headers=self.headers
        )
        self.app.post_json(
            "/buckets/test/collections/alice-julia/records",
            {"permissions": {"read": ["system.Authenticated"]}},
            headers=self.headers,
        )

    def test_author_can_read_everything(self):
        resp = self.app.get("/buckets/test/history", headers=self.headers)
        entries = resp.json["data"]
        assert len(entries) == 6  # everything.

    def test_read_permission_can_be_given_to_anybody_via_settings(self):
        with mock.patch.dict(
            self.app.app.registry.settings, [("history_read_principals", "system.Everyone")]
        ):
            resp = self.app.get("/buckets/test/history", headers=get_user_headers("tartan:pion"))
            entries = resp.json["data"]
            assert len(entries) == 6  # everything.

    def test_bucket_read_allows_whole_history(self):
        resp = self.app.get("/buckets/test/history", headers=self.alice_headers)
        entries = resp.json["data"]
        assert len(entries) == 6  # everything.

        self.app.get("/buckets/author-only/history", headers=self.alice_headers, status=403)

    def test_collection_read_restricts_to_collection(self):
        resp = self.app.get("/buckets/test/history", headers=self.julia_headers)
        entries = resp.json["data"]
        assert len(entries) == 3
        assert entries[0]["resource_name"] == "record"
        assert entries[1]["resource_name"] == "record"
        assert entries[2]["resource_name"] == "collection"

    def test_write_on_record_restricts_to_record(self):
        resp = self.app.get("/buckets/test/history", headers=self.bob_headers)
        entries = resp.json["data"]
        assert len(entries) == 2
        assert "system.Authenticated" in entries[0]["target"]["permissions"]["read"]
        assert entries[0]["resource_name"] == "record"
        assert self.bob_principal in entries[1]["target"]["permissions"]["write"]
        assert entries[1]["resource_name"] == "record"

    def test_publicly_readable_record_allows_any_authenticated(self):
        resp = self.app.get("/buckets/test/history", headers=get_user_headers("jack:"))
        entries = resp.json["data"]
        assert len(entries) == 1
        assert "system.Authenticated" in entries[0]["target"]["permissions"]["read"]
        assert entries[0]["resource_name"] == "record"

    def test_new_entries_are_not_readable_if_permission_is_removed(self):
        resp = self.app.get("/buckets/test/history", headers=self.alice_headers)
        before = resp.headers["ETag"]

        # Remove alice from read permission.
        self.app.patch_json("/buckets/test", {"permissions": {"read": []}}, headers=self.headers)

        # Create new collection.
        self.app.put_json("/buckets/test/collections/new-one", headers=self.headers)

        # History did not evolve for alice.
        resp = self.app.get("/buckets/test/history", headers=self.alice_headers)
        assert resp.headers["ETag"] != before

    def test_history_entries_are_not_listed_in_permissions_endpoint(self):
        resp = self.app.get("/permissions", headers=self.headers)
        entries = [e["resource_name"] == "history" for e in resp.json["data"]]
        assert not any(entries)


class ExcludeResourcesTest(HistoryWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["history.exclude_resources"] = (
            "/buckets/a " "/buckets/b/collections/a " "/buckets/b/groups/a"
        )
        return settings

    def setUp(self):
        group = {"data": {"members": []}}
        self.app.put_json("/buckets/a", headers=self.headers)
        self.app.put_json("/buckets/a/groups/admins", group, headers=self.headers)
        self.app.put_json("/buckets/b", headers=self.headers)
        self.app.put_json("/buckets/b/groups/a", group, headers=self.headers)
        self.app.put_json("/buckets/b/collections/a", headers=self.headers)
        self.app.put_json("/buckets/b/collections/a/records/1", headers=self.headers)
        self.app.put_json("/buckets/b/collections/b", headers=self.headers)
        self.app.put_json("/buckets/b/collections/b/records/1", headers=self.headers)

    def test_whole_buckets_can_be_excluded(self):
        resp = self.app.get("/buckets/a/history", headers=self.headers)
        entries = resp.json["data"]
        assert len(entries) == 0  # nothing.

    def test_some_specific_collection_can_be_excluded(self):
        resp = self.app.get("/buckets/b/history?collection_id=b", headers=self.headers)
        entries = resp.json["data"]
        assert len(entries) > 0

        resp = self.app.get("/buckets/b/history?collection_id=a", headers=self.headers)
        entries = resp.json["data"]
        assert len(entries) == 0  # nothing.

    def test_some_specific_object_can_be_excluded(self):
        resp = self.app.get("/buckets/b/history?group_id=a", headers=self.headers)
        entries = resp.json["data"]
        assert len(entries) == 0  # nothing.


class DisabledExplicitPermissionsTest(HistoryWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["explicit_permissions"] = "false"
        return settings

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.alice_headers = {**cls.headers, **get_user_headers("alice")}
        cls.alice_principal = (
            "basicauth:d5b0026601f1b251974e09548d44155e16812e3c64ff7ae053fe3542e2ca1570"
        )

    def setUp(self):
        self.app.put_json(
            "/buckets/test",
            {"permissions": {"write": ["system.Authenticated"]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/test/collections/test",
            {"permissions": {"write": ["system.Authenticated"]}},
            headers=self.headers,
        )

    def test_history_can_still_be_read(self):
        self.app.post_json("/buckets/test/collections/test/records", headers=self.alice_headers)

        resp = self.app.get("/buckets/test/history", headers=self.alice_headers)
        self.assertEqual(
            [(entry["action"], entry["resource_name"]) for entry in resp.json["data"]],
            [("create", "record"), ("create", "collection"), ("create", "bucket")],
        )



import re
import uuid

import colander
from pyramid import httpexceptions
from pyramid.authorization import Authenticated, Everyone
from pyramid.decorator import reify
from pyramid.events import subscriber
from pyramid.settings import aslist

from kinto.core import resource
from kinto.core.errors import http_error, raise_invalid
from kinto.core.events import ACTIONS, ResourceChanged
from kinto.views import NameGenerator

from ..mails import Emailer
from ..utils import (
    ACCOUNT_POLICY_NAME,
    cache_validation_key,
    delete_cached_account,
    get_cached_validation_key,
    hash_password,
)


DEFAULT_EMAIL_REGEXP = "^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$"


def _extract_posted_body_id(request):
    try:
        # Anonymous creation with POST.
        return request.json["data"]["id"]
    except (ValueError, KeyError):
        # Bad POST data.
        if request.method.lower() == "post":
            error_details = {"name": "data.id", "description": "data.id in body: Required"}
            raise_invalid(request, **error_details)
        # Anonymous GET
        error_msg = "Cannot read accounts."
        raise http_error(httpexceptions.HTTPUnauthorized(), error=error_msg)


class AccountIdGenerator(NameGenerator):
    """Allow @ signs in account IDs."""

    regexp = r"^[a-zA-Z0-9][+.@a-zA-Z0-9_-]*$"


class AccountSchema(resource.ResourceSchema):
    password = colander.SchemaNode(colander.String())


@resource.register()
class Account(resource.Resource):
    schema = AccountSchema

    def __init__(self, request, context):
        settings = request.registry.settings
        # Store if current user is administrator (before accessing get_parent_id())
        allowed_from_settings = settings.get("account_write_principals", [])
        context.is_administrator = (
            len(set(aslist(allowed_from_settings)) & set(request.prefixed_principals)) > 0
        )
        # Shortcut to check if current is anonymous (before get_parent_id()).
        context.is_anonymous = Authenticated not in request.effective_principals
        # Is the "accounts validation" setting set?
        context.validation_enabled = settings.get("account_validation", False)
        # Account validation requires the user id to be an email.
        validation_email_regexp = settings.get(
            "account_validation.email_regexp", DEFAULT_EMAIL_REGEXP
        )
        context.validation_email_regexp = re.compile(validation_email_regexp)

        super().__init__(request, context)

        # Overwrite the current principal set by Resource.
        if self.model.current_principal == Everyone or context.is_administrator:
            # Creation is anonymous, but author with write perm is this:
            self.model.current_principal = f"{ACCOUNT_POLICY_NAME}:{self.model.parent_id}"

    @reify
    def id_generator(self):
        # This generator is used for ID validation.
        return AccountIdGenerator()

    def get_parent_id(self, request):
        # The whole challenge here is that we want to isolate what
        # authenticated users can list, but give access to everything to
        # administrators.
        # Plus when anonymous create accounts, we have to set their parent id
        # to the same value they would obtain when authenticated.
        if self.context.is_administrator:
            if self.context.on_plural_endpoint:
                # Accounts created by admin should have userid as parent.
                if request.method.lower() == "post":
                    return _extract_posted_body_id(request)
                else:
                    # Admin see all accounts.
                    return "*"
            else:
                # No pattern matching for admin on single record.
                return request.matchdict["id"]

        if not self.context.is_anonymous:
            # Authenticated users see their own account only.
            return request.selected_userid

        # Anonymous creation with PUT.
        if "id" in request.matchdict:
            return request.matchdict["id"]

        return _extract_posted_body_id(request)

    def process_object(self, new, old=None):
        new = super(Account, self).process_object(new, old)

        if "data" in self.request.json and "password" in self.request.json["data"]:
            new["password"] = hash_password(new["password"])

        # Do not let accounts be created without usernames.
        if self.model.id_field not in new:
            error_details = {"name": "data.id", "description": "Accounts must have an ID."}
            raise_invalid(self.request, **error_details)

        # Account validation requires that the record ID is an email address.
        # TODO: this might be better suited for a schema. Do we have a way to
        # dynamically change the schema according to the settings?
        if self.context.validation_enabled and old is None:
            email_regexp = self.context.validation_email_regexp
            # Account validation requires that the record ID is an email address.
            user_email = new[self.model.id_field]
            if not email_regexp.match(user_email):
                error_details = {
                    "name": "data.id",
                    "description": f"Account validation is enabled, and user id should match {email_regexp}",
                }
                raise_invalid(self.request, **error_details)

            activation_key = str(uuid.uuid4())
            new["validated"] = False

            # Store the activation key in the cache to be used in the `validate` endpoint.
            cache_validation_key(activation_key, new["id"], self.request.registry)

        # Administrators can reach other accounts and anonymous have no
        # selected_userid. So do not try to enforce.
        if self.context.is_administrator or self.context.is_anonymous:
            return new

        # Otherwise, we force the id to match the authenticated username.
        if new[self.model.id_field] != self.request.selected_userid:
            error_details = {
                "name": "data.id",
                "description": "Username and account ID do not match.",
            }
            raise_invalid(self.request, **error_details)

        return new


# Clear cache on account change
@subscriber(
    ResourceChanged, for_resources=("account",), for_actions=(ACTIONS.UPDATE, ACTIONS.DELETE)
)
def on_account_changed(event):
    request = event.request

    for obj in event.impacted_objects:
        # Extract username and password from current user
        username = obj["old"]["id"]
        # Delete cache
        delete_cached_account(username, request.registry)


# Send activation code by email on account creation if account validation is enabled.
@subscriber(ResourceChanged, for_resources=("account",), for_actions=(ACTIONS.CREATE,))
def on_account_created(event):
    request = event.request
    settings = request.registry.settings
    if not settings.get("account_validation", False):
        return

    for impacted_object in event.impacted_objects:
        account = impacted_object["new"]
        user_email = account["id"]
        activation_key = get_cached_validation_key(user_email, request.registry)
        if activation_key is None:
            continue

        # Send an email to the user with the link to activate their account.
        Emailer(request, account).send_activation(activation_key)



from bravado_core.request import IncomingRequest, unmarshal_request
from bravado_core.swagger20_validator import ValidationError

from .support import OpenAPITest


class OpenAPIRequestsValidationTest(OpenAPITest):
    def setUp(self):
        super().setUp()

        self.request = IncomingRequest()
        self.request.path = {}
        self.request.headers = {}
        self.request.query = {}
        self.request._json = {}
        self.request.json = lambda: self.request._json

    def test_validate_bucket_path(self):
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Buckets"].get_bucket
        )

    def test_validate_groups_path(self):
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Groups"].get_groups
        )

    def test_validate_group_path(self):
        paths = [{}, {"bucket_id": "b1"}, {"id": "g1"}]
        for path in paths:
            self.request.path = path
            self.assertRaises(
                ValidationError,
                unmarshal_request,
                self.request,
                self.resources["Groups"].get_group,
            )

    def test_validate_collections_path(self):
        self.assertRaises(
            ValidationError,
            unmarshal_request,
            self.request,
            self.resources["Collections"].get_collections,
        )

    def test_validate_plural_path(self):
        paths = [{}, {"bucket_id": "b1"}, {"id": "c1"}]
        for path in paths:
            self.request.path = path
            self.assertRaises(
                ValidationError,
                unmarshal_request,
                self.request,
                self.resources["Collections"].get_collection,
            )

    def test_validate_records_path(self):
        paths = [{}, {"bucket_id": "b1"}, {"collection_id": "c1"}]
        for path in paths:
            self.request.path = path
            self.assertRaises(
                ValidationError,
                unmarshal_request,
                self.request,
                self.resources["Records"].get_records,
            )

    def test_validate_object_path(self):
        paths = [{}, {"bucket_id": "b1", "collection_id": "c1"}, {"id": "r1"}]
        for path in paths:
            self.request.path = path
            self.assertRaises(
                ValidationError,
                unmarshal_request,
                self.request,
                self.resources["Records"].get_record,
            )

    def test_validate_data(self):
        bodies = [{"data": "aaa"}]
        for body in bodies:
            self.request._json = body
            self.assertRaises(
                ValidationError,
                unmarshal_request,
                self.request,
                self.resources["Buckets"].create_bucket,
            )

    def test_validate_permissions(self):
        bodies = [
            {"permissions": "aaa"},
            {"permissions": {"read": "aaa"}},
            {"permissions": {"read": [111]}},
        ]
        for body in bodies:
            self.request._json = body
            self.assertRaises(
                ValidationError,
                unmarshal_request,
                self.request,
                self.resources["Buckets"].create_bucket,
            )

    def test_validate_queries(self):
        queries = [{"_since": "aaa"}, {"_before": "aaa"}, {"_limit": "aaa"}, {"_token": {}}]
        for query in queries:
            self.request.query = query
            self.assertRaises(
                ValidationError,
                unmarshal_request,
                self.request,
                self.resources["Buckets"].get_buckets,
            )

    def test_validate_headers(self):
        headers = [{"If-None-Match": "123"}, {"If-Match": "123"}]
        for head in headers:
            self.request.headers = head
            self.assertRaises(
                ValidationError,
                unmarshal_request,
                self.request,
                self.resources["Buckets"].get_buckets,
            )

    def test_validate_batch_requests_method(self):
        self.request._json = {"requests": [{"method": "AAA", "path": "/buckets/b1"}]}
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Batch"].batch
        )

    def test_validate_batch_requests_path(self):
        self.request._json = {"requests": [{"method": "GET", "path": 123}]}
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Batch"].batch
        )

    def test_validate_batch_requests_body(self):
        self.request._json = {"requests": [{"method": "GET", "path": "/buckets/b1", "body": []}]}
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Batch"].batch
        )

    def test_validate_batch_requests_header(self):
        self.request._json = {
            "requests": [{"method": "GET", "path": "/buckets/b1", "body": {}, "headers": []}]
        }
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Batch"].batch
        )

    def test_validate_batch_defaults(self):
        self.request._json = {
            "defaults": [],
            "requests": [{"method": "GET", "path": "/buckets/b1"}],
        }
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Batch"].batch
        )

    def test_validate_batch_defaults_method(self):
        self.request._json = {"defaults": {"method": "AAA"}, "requests": [{"path": "/buckets/b1"}]}
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Batch"].batch
        )

    def test_validate_batch_defaults_body(self):
        self.request._json = {
            "defaults": {"body": []},
            "requests": [{"method": "PUT", "path": "/buckets/b1"}],
        }
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Batch"].batch
        )

    def test_validate_batch_defaults_headers(self):
        self.request._json = {
            "defaults": {"headers": []},
            "requests": [{"method": "GET", "path": "/buckets/b1"}],
        }
        self.assertRaises(
            ValidationError, unmarshal_request, self.request, self.resources["Batch"].batch
        )



import requests


_configs = {}


def fetch_openid_config(issuer):
    global _configs

    if issuer not in _configs:
        resp = requests.get(issuer.rstrip("/") + "/.well-known/openid-configuration")
        _configs[issuer] = resp.json()

    return _configs[issuer]



import unittest
from unittest import mock

from kinto import scripts


class RebuildQuotasTest(unittest.TestCase):
    def setUp(self):
        self.registry = mock.MagicMock()
        self.registry.settings = {"includes": "kinto.plugins.quotas"}

    def test_rebuild_quotas_in_read_only_display_an_error(self):
        with mock.patch("kinto.scripts.logger") as mocked:
            self.registry.settings["readonly"] = "true"
            code = scripts.rebuild_quotas({"registry": self.registry})
            assert code == 41
            mocked.error.assert_any_call("Cannot rebuild quotas while " "in readonly mode.")

    def test_rebuild_quotas_when_not_included_display_an_error(self):
        with mock.patch("kinto.scripts.logger") as mocked:
            self.registry.settings["includes"] = ""
            code = scripts.rebuild_quotas({"registry": self.registry})
            assert code == 42
            mocked.error.assert_any_call(
                "Cannot rebuild quotas when " "quotas plugin is not installed."
            )

    def test_rebuild_quotas_calls_quotas_script(self):
        with mock.patch("kinto.scripts.quotas.rebuild_quotas") as mocked:
            code = scripts.rebuild_quotas({"registry": self.registry})
            assert code == 0
            mocked.assert_called_with(self.registry.storage, dry_run=False)



import json


def record_size(record):
    # We cannot use rapidjson here, since the `separator` option is not available.
    canonical_json = json.dumps(record, sort_keys=True, separators=(",", ":"))
    return len(canonical_json)



import time
from unittest import mock

import pytest
from pyramid import testing

from kinto.core.cache import heartbeat
from kinto.core.storage import exceptions


class CacheTest:
    backend = None
    settings = {}

    def setUp(self):
        super().setUp()
        self.cache = self.backend.load_from_config(self._get_config())
        self.cache.initialize_schema()
        self.request = None
        self.client_error_patcher = None

    def _get_config(self, settings=None):
        """Mock Pyramid config object."""
        if settings is None:
            settings = self.settings
        config = testing.setUp()
        config.add_settings(settings)
        return config

    def tearDown(self):
        mock.patch.stopall()
        super().tearDown()
        self.cache.flush()

    def get_backend_prefix(self, prefix):
        settings_prefix = {**self.settings}
        settings_prefix["cache_prefix"] = prefix
        config_prefix = self._get_config(settings=settings_prefix)

        # initiating cache backend with prefix:
        backend_prefix = self.backend.load_from_config(config_prefix)

        return backend_prefix

    def test_backend_error_is_raised_anywhere(self):
        self.client_error_patcher.start()
        calls = [
            (self.cache.flush,),
            (self.cache.ttl, ""),
            (self.cache.expire, "", 0),
            (self.cache.get, ""),
            (self.cache.set, "", "", 42),
            (self.cache.delete, ""),
        ]
        for call in calls:
            self.assertRaises(exceptions.BackendError, *call)

    def test_initialize_schema_is_idempotent(self):
        self.cache.initialize_schema()
        self.cache.initialize_schema()  # not raising.

    def test_ping_returns_false_if_unavailable(self):
        self.client_error_patcher.start()
        ping = heartbeat(self.cache)
        self.assertFalse(ping(self.request))
        with mock.patch("kinto.core.cache.random.SystemRandom.random", return_value=0.6):
            self.assertFalse(ping(self.request))
        with mock.patch("kinto.core.cache.random.SystemRandom.random", return_value=0.4):
            self.assertFalse(ping(self.request))

    def test_ping_returns_true_if_available(self):
        ping = heartbeat(self.cache)
        with mock.patch("kinto.core.cache.random.random", return_value=0.6):
            self.assertTrue(ping(self.request))
        with mock.patch("kinto.core.cache.random.random", return_value=0.4):
            self.assertTrue(ping(self.request))

    def test_ping_logs_error_if_unavailable(self):
        self.client_error_patcher.start()
        ping = heartbeat(self.cache)

        with mock.patch("kinto.core.cache.logger.exception") as exc_handler:
            self.assertFalse(ping(self.request))

        self.assertTrue(exc_handler.called)

    def test_set_adds_the_object(self):
        stored = "toto"
        self.cache.set("foobar", stored, 42)
        retrieved = self.cache.get("foobar")
        self.assertEqual(retrieved, stored)

    def test_values_remains_python_dict(self):
        def setget(k, v):
            self.cache.set(k, v, 42)
            return (self.cache.get(k), v)

        self.assertEqual(*setget("foobar", 3))
        self.assertEqual(*setget("foobar", ["a"]))
        self.assertEqual(*setget("foobar", {"b": [1, 2]}))
        self.assertEqual(*setget("foobar", 3.14))

    def test_bytes_cannot_be_stored_in_the_cache(self):
        with pytest.raises(TypeError):
            self.cache.set("test", b"foo", 42)

    def test_delete_removes_the_object(self):
        self.cache.set("foobar", "toto", 42)
        returned = self.cache.delete("foobar")
        self.assertEqual(returned, "toto")
        missing = self.cache.get("foobar")
        self.assertIsNone(missing)

    def test_delete_does_not_fail_if_object_is_unknown(self):
        returned = self.cache.delete("foobar")
        self.assertIsNone(returned)

    def test_expire_expires_the_value(self):
        self.cache.set("foobar", "toto", 42)
        self.cache.expire("foobar", 0.01)
        time.sleep(0.02)
        retrieved = self.cache.get("foobar")
        self.assertIsNone(retrieved)

    def test_set_with_ttl_expires_the_value(self):
        self.cache.set("foobar", "toto", 0.01)
        time.sleep(0.02)
        retrieved = self.cache.get("foobar")
        self.assertIsNone(retrieved)

    def test_ttl_return_the_time_to_live(self):
        self.cache.set("foobar", "toto", 42)
        self.cache.expire("foobar", 10)
        ttl = self.cache.ttl("foobar")
        self.assertGreater(ttl, 0)
        self.assertLessEqual(ttl, 10)

    def test_ttl_return_none_if_unknown(self):
        ttl = self.cache.ttl("unknown")
        self.assertTrue(ttl < 0)

    def test_cache_prefix_is_set(self):
        backend_prefix = self.get_backend_prefix(prefix="prefix_")

        # Set the value
        backend_prefix.set("key", "foo", 42)

        # Validate that it was set with the prefix.
        obtained = self.cache.get("prefix_key")
        self.assertEqual(obtained, "foo")

    def test_cache_when_prefix_is_not_set(self):
        backend_prefix = self.get_backend_prefix(prefix="")

        # Set a value
        backend_prefix.set("key", "foo", 42)

        # Validate that it was set with no prefix
        obtained = self.cache.get("key")
        self.assertEqual(obtained, "foo")

    def test_prefix_value_use_to_get_data(self):
        backend_prefix = self.get_backend_prefix(prefix="prefix_")

        # Set the value with the prefix
        self.cache.set("prefix_key", "foo", 42)

        # Validate that the prefix was added
        obtained = backend_prefix.get("key")
        self.assertEqual(obtained, "foo")

    def test_prefix_value_use_to_delete_data(self):
        backend_prefix = self.get_backend_prefix(prefix="prefix_")
        # Set the value
        self.cache.set("prefix_key", "foo", 42)

        # Delete the value
        backend_prefix.delete("key")

        # Validate that the value was deleted
        obtained = self.cache.get("prefix_key")
        self.assertEqual(obtained, None)

    def test_prefix_value_used_with_ttl(self):
        backend_prefix = self.get_backend_prefix(prefix="prefix_")

        self.cache.set("prefix_key", "foo", 10)

        # Validate that the ttl add the prefix to the key.
        obtained = backend_prefix.ttl("key")
        self.assertLessEqual(obtained, 10)
        self.assertGreater(obtained, 9)

    def test_prefix_value_used_with_expire(self):
        backend_prefix = self.get_backend_prefix(prefix="prefix_")

        self.cache.set("prefix_foobar", "toto", 10)

        # expiring the ttl of key
        backend_prefix.expire("foobar", 0)

        # Make sure the TTL was set accordingly.
        ttl = self.cache.ttl("prefix_foobar")
        self.assertLessEqual(ttl, 0)

        # The object should have expired
        retrieved = self.cache.get("prefix_foobar")
        self.assertIsNone(retrieved)



from kinto import __version__ as VERSION
from kinto.core.testing import unittest

from .support import MINIMALIST_BUCKET, MINIMALIST_GROUP, BaseWebTest


class HelloViewTest(BaseWebTest, unittest.TestCase):
    def test_returns_info_about_url_and_version(self):
        response = self.app.get("/")
        self.assertEqual(response.json["project_name"], "kinto")
        self.assertEqual(response.json["project_version"], VERSION)
        self.assertEqual(response.json["project_docs"], "https://kinto.readthedocs.io/")
        self.assertEqual(response.json["url"], "http://localhost/v1/")

    def test_project_name_can_be_overriden(self):
        settings = {"project_name": "RemoteSettings"}
        app = self.make_app(settings=settings)
        resp = app.get("/")
        self.assertEqual(resp.json["project_name"], "RemoteSettings")

    def test_hides_user_info_if_anonymous(self):
        response = self.app.get("/")
        self.assertNotIn("user", response.json)

    def test_cache_control_if_readonly(self):
        settings = self.get_app_settings()
        settings["readonly"] = True
        app = self.make_app(settings=settings)
        response = app.get("/")
        self.assertIn("Cache-Control", response.headers)
        self.assertEqual(response.headers["Cache-Control"], "max-age=86400")

    def test_no_cache_control_if_not_readonly(self):
        response = self.app.get("/", headers=self.headers)
        self.assertNotIn("Cache-Control", response.headers)

    def test_returns_user_id_if_authenticated(self):
        response = self.app.get("/", headers=self.headers)
        self.assertEqual(response.json["user"]["id"], self.principal)

    def test_returns_user_principals_if_authenticated(self):
        group_url = "/buckets/beers/groups/users"
        group = {**MINIMALIST_GROUP}
        group["data"]["members"].append(self.principal)
        self.app.put_json("/buckets/beers", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(group_url, group, headers=self.headers)
        response = self.app.get("/", headers=self.headers).json["user"]["principals"]
        principals = ("system.Everyone", "system.Authenticated", group_url, self.principal)
        self.assertEqual(sorted(response), sorted(principals))

    def test_capability_is_exposed_if_setting_is_set(self):
        settings = {"experimental_collection_schema_validation": True}
        app = self.make_app(settings=settings)
        resp = app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("schema", capabilities)
        expected = {
            "description": "Validates collection records with JSON schemas.",
            "url": "https://kinto.readthedocs.io/en/latest/api/1.x/"
            "collections.html#collection-json-schema",
        }
        self.assertEqual(expected, capabilities["schema"])

    def test_capability_is_exposed_if_setting_is_not_set(self):
        settings = self.get_app_settings()
        settings["experimental_collection_schema_validation"] = False
        app = self.make_app(settings=settings)
        resp = app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertNotIn("schema", capabilities)

    def test_permissions_capability_if_enabled(self):
        settings = {"experimental_permissions_endpoint": True}
        app = self.make_app(settings=settings)
        resp = app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("permissions_endpoint", capabilities)
        expected = {
            "description": "The permissions endpoint can be used to list "
            "all user objects permissions.",
            "url": "https://kinto.readthedocs.io/en/latest/configuration/"
            "settings.html#activating-the-permissions-endpoint",
        }
        self.assertEqual(expected, capabilities["permissions_endpoint"])

    def test_permissions_capability_if_not_enabled(self):
        settings = {"experimental_permissions_endpoint": False}
        app = self.make_app(settings=settings)
        resp = app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertNotIn("permissions_endpoint", capabilities)



import unittest

from pyramid.config import Configurator

from kinto.core.testing import get_user_headers
from kinto.events import ServerFlushed

from ..support import MINIMALIST_BUCKET, MINIMALIST_COLLECTION, MINIMALIST_RECORD, BaseWebTest


class FlushViewTest(BaseWebTest, unittest.TestCase):
    collection_url = "/buckets/beers/collections/barley/records"
    events = []

    def setUp(self):
        super().setUp()

        del self.events[:]

        bucket = {**MINIMALIST_BUCKET}

        self.alice_headers = {**self.headers, **get_user_headers("alice")}

        resp = self.app.get("/", headers=self.alice_headers)
        alice_principal = resp.json["user"]["id"]
        bucket["permissions"] = {"write": [alice_principal]}

        # Create shared bucket.
        self.app.put_json("/buckets/beers", bucket, headers=self.headers)
        self.app.put_json(
            "/buckets/beers/collections/barley", MINIMALIST_COLLECTION, headers=self.headers
        )

        # Records for alice and bob.
        self.app.post_json(
            self.collection_url, MINIMALIST_RECORD, headers=self.headers, status=201
        )
        self.app.post_json(
            self.collection_url, MINIMALIST_RECORD, headers=self.alice_headers, status=201
        )

    def tearDown(self):
        del self.events[:]
        super().tearDown()

    @classmethod
    def make_app(cls, settings=None, config=None):
        settings = cls.get_app_settings(settings)
        config = Configurator(settings=settings)
        config.add_subscriber(cls.listener, ServerFlushed)
        config.commit()
        return super().make_app(settings=settings, config=config)

    @classmethod
    def get_app_settings(cls, extras=None):
        if extras is None:
            extras = {}
        extras.setdefault("includes", "kinto.plugins.flush")
        settings = super().get_app_settings(extras)
        return settings

    @classmethod
    def listener(cls, event):
        cls.events.append(event)

    def test_returns_404_if_not_enabled_in_configuration(self):
        app = self.make_app(settings={"includes": ""})
        app.post("/__flush__", headers=self.headers, status=404)

    def test_removes_every_records_of_everykind(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.app.get(self.collection_url, headers=self.alice_headers)

        self.app.post("/__flush__", headers=self.headers, status=202)

        self.app.get(self.collection_url, headers=self.headers, status=403)
        self.app.get(self.collection_url, headers=self.alice_headers, status=403)

    def test_event_triggered_post(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.app.get(self.collection_url, headers=self.alice_headers)
        self.app.post("/__flush__", headers=self.headers, status=202)
        self.assertEqual(len(self.events), 1)
        self.assertTrue(isinstance(self.events[0], ServerFlushed))

    def test_flush_returns_json(self):
        response = self.app.post("/__flush__", headers=self.headers, status=202)
        self.assertEqual(response.json, {})

    def test_flush_capability_if_enabled(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("flush_endpoint", capabilities)
        expected = {
            "description": "The __flush__ endpoint can be used to remove "
            "all data from all backends.",
            "url": "https://kinto.readthedocs.io/en/latest/api/1.x/flush.html",
        }
        self.assertEqual(expected, capabilities["flush_endpoint"])



from unittest import mock

import colander

from kinto.core.resource import schema
from kinto.core.testing import unittest


class DepracatedSchemasTest(unittest.TestCase):
    def test_resource_timestamp_is_depracated(self):
        with mock.patch("kinto.core.resource.schema.warnings") as mocked:
            schema.TimeStamp()
            message = (
                "`kinto.core.resource.schema.TimeStamp` is deprecated, "
                "use `kinto.core.schema.TimeStamp` instead."
            )
            mocked.warn.assert_called_with(message, DeprecationWarning)

    def test_resource_URL_is_depracated(self):
        with mock.patch("kinto.core.resource.schema.warnings") as mocked:
            schema.URL()
            message = (
                "`kinto.core.resource.schema.URL` is deprecated, "
                "use `kinto.core.schema.URL` instead."
            )
            mocked.warn.assert_called_with(message, DeprecationWarning)


class ResourceSchemaTest(unittest.TestCase):
    def test_preserves_unknown_fields_when_specified(self):
        class PreserveSchema(schema.ResourceSchema):
            class Options:
                preserve_unknown = True

        schema_instance = PreserveSchema()
        deserialized = schema_instance.deserialize({"foo": "bar"})
        self.assertIn("foo", deserialized)
        self.assertEqual(deserialized["foo"], "bar")

    def test_ignore_unknwon_fields_when_specified(self):
        class PreserveSchema(schema.ResourceSchema):
            class Options:
                preserve_unknown = False

        schema_instance = PreserveSchema()
        deserialized = schema_instance.deserialize({"foo": "bar"})
        self.assertNotIn("foo", deserialized)

    def test_accepts_unknown_fields_by_default(self):
        schema_instance = schema.ResourceSchema()
        deserialized = schema_instance.deserialize({"foo": "bar"})
        self.assertIn("foo", deserialized)

    def test_options_parameters_use_default_value_when_subclassed(self):
        class PreserveSchema(schema.ResourceSchema):
            class Options:
                pass

        schema_instance = PreserveSchema()
        deserialized = schema_instance.deserialize({"foo": "bar"})
        self.assertIn("foo", deserialized)


class PermissionsSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = schema.PermissionsSchema()

    def test_works_with_any_permission_name_by_default(self):
        perms = {"can_cook": ["mat"]}
        deserialized = self.schema.deserialize(perms)
        self.assertEqual(deserialized, perms)

    def test_works_with_empty_mapping(self):
        perms = {}
        deserialized = self.schema.deserialize(perms)
        self.assertEqual(deserialized, perms)

    def test_works_with_empty_list_of_principals(self):
        perms = {"can_cook": []}
        deserialized = self.schema.deserialize(perms)
        self.assertEqual(deserialized, perms)

    def test_works_with_empty_none_permissions(self):
        perms = {"can_cook": None}
        deserialized = self.schema.deserialize(perms)
        self.assertEqual(deserialized, perms)

    def test_raises_invalid_if_not_a_mapping(self):
        perms = ["gab"]
        self.assertRaises(colander.Invalid, self.schema.deserialize, perms)

    def test_raises_invalid_if_permission_is_unknown(self):
        self.schema.known_perms = ("can_sleep",)
        perms = {"can_work": ["mat"]}
        self.assertRaises(colander.Invalid, self.schema.deserialize, perms)

    def test_raises_invalid_if_not_list(self):
        perms = {"can_cook": 3.14}
        self.assertRaises(colander.Invalid, self.schema.deserialize, perms)

    def test_raises_invalid_if_not_list_of_strings(self):
        perms = {"can_cook": ["pi", 3.14]}
        self.assertRaises(colander.Invalid, self.schema.deserialize, perms)


class HeaderFieldSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = schema.HeaderField(colander.String())

    def test_decode_unicode(self):
        value = "\xe7 is not a c"
        deserialized = self.schema.deserialize(value.encode("utf-8"))
        self.assertEqual(deserialized, value)

    def test_bad_unicode_raises_invalid(self):
        value = b"utf8 \xe9"
        self.assertRaises(colander.Invalid, self.schema.deserialize, value)


class QueryFieldSchemaTest(unittest.TestCase):
    def test_deserialize_integer_between_quotes(self):
        self.schema = schema.QueryField(colander.Integer())
        deserialized = self.schema.deserialize("123")
        self.assertEqual(deserialized, 123)


class FieldListSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = schema.FieldList()

    def test_deserialize_as_list(self):
        value = "foo,-bar,123"
        deserialized = self.schema.deserialize(value)
        self.assertEqual(deserialized, ["foo", "-bar", "123"])

    def test_handle_drop(self):
        value = colander.null
        deserialized = self.schema.deserialize(value)
        self.assertEqual(deserialized, colander.drop)

    def test_handle_empty(self):
        value = ""
        deserialized = self.schema.deserialize(value)
        self.assertEqual(deserialized, [])

    def test_raises_invalid_if_not_string(self):
        value = 123
        self.assertRaises(colander.Invalid, self.schema.deserialize, value)


class HeaderQuotedIntegerSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = schema.HeaderQuotedInteger()

    def test_deserialize_as_integer(self):
        value = '"123"'
        deserialized = self.schema.deserialize(value)
        self.assertEqual(deserialized, 123)

    def test_deserialize_any(self):
        value = "*"
        deserialized = self.schema.deserialize(value)
        self.assertEqual(deserialized, "*")

    def test_unquoted_raises_invalid(self):
        value = "123"
        self.assertRaises(colander.Invalid, self.schema.deserialize, value)

    def test_integer_raises_invalid(self):
        value = 123
        self.assertRaises(colander.Invalid, self.schema.deserialize, value)

    def test_invalid_quoted_raises_invalid(self):
        value = "foo"
        self.assertRaises(colander.Invalid, self.schema.deserialize, value)

    def test_empty_raises_invalid(self):
        value = '""'
        self.assertRaises(colander.Invalid, self.schema.deserialize, value)


class ObjectSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = schema.ObjectSchema()

    def test_binds_data(self):
        bound = self.schema.bind(data=schema.ResourceSchema())
        value = {"data": {"foo": "bar"}}
        deserialized = bound.deserialize(value)
        self.assertEqual(deserialized, value)

    def test_binds_permissions(self):
        permissions = schema.PermissionsSchema(permissions=("sleep",))
        bound = self.schema.bind(permissions=permissions)
        value = {"permissions": {"sleep": []}}
        deserialized = bound.deserialize(value)
        self.assertEqual(deserialized, value)

    def test_allow_binding_perms_after_data(self):
        bound = self.schema.bind(data=schema.ResourceSchema())
        permissions = schema.PermissionsSchema(permissions=("sleep",))
        bound = bound.bind(permissions=permissions)
        value = {"data": {"foo": "bar"}, "permissions": {"sleep": []}}
        deserialized = bound.deserialize(value)
        self.assertEqual(deserialized, value)

    def test_doesnt_allow_permissions_unless_bound(self):
        bound = self.schema.bind(data=schema.ResourceSchema())
        value = {"permissions": {"sleep": []}}
        self.assertRaises(colander.UnsupportedFields, bound.deserialize, value)


class RequestSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = schema.RequestSchema()

    def test_header_supports_binding(self):
        header = colander.MappingSchema(missing={"foo": "bar"})
        bound = self.schema.bind(header=header)
        deserialized = bound.deserialize({})
        self.assertEqual(deserialized["header"], {"foo": "bar"})

    def test_querystring_supports_binding(self):
        querystring = colander.MappingSchema(missing={"foo": "bar"})
        bound = self.schema.bind(querystring=querystring)
        deserialized = bound.deserialize({})
        self.assertEqual(deserialized["querystring"], {"foo": "bar"})

    def test_default_header_if_not_bound(self):
        bound = self.schema.bind()
        self.assertEqual(type(bound["header"]), schema.HeaderSchema)

    def test_default_querystring_if_not_bound(self):
        bound = self.schema.bind()
        self.assertEqual(type(bound["querystring"]), schema.QuerySchema)

    def test_header_preserve_unknown_fields(self):
        value = {"header": {"foo": "bar"}}
        deserialized = self.schema.bind().deserialize(value)
        self.assertEqual(deserialized, value)

    def test_querystring_preserve_unknown_fields(self):
        value = {"querystring": {"foo": "bar"}}
        deserialized = self.schema.bind().deserialize(value)
        self.assertEqual(deserialized, value)

    def test_drops(self):
        deserialized = self.schema.bind().deserialize({})
        self.assertEqual(deserialized, {})


class PayloadRequestSchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = schema.PayloadRequestSchema()

    def test_body_supports_binding(self):
        body = colander.MappingSchema(missing={"foo": "bar"})
        bound = self.schema.bind(body=body)
        deserialized = bound.deserialize({})
        self.assertEqual(deserialized["body"], {"foo": "bar"})

    def test_body_supports_binding_after_other_binds(self):
        querystring = colander.MappingSchema(missing={"foo": "bar"})
        bound = self.schema.bind(querystring=querystring)
        body = colander.MappingSchema(missing={"foo": "beer"})
        bound = bound.bind(body=body)
        deserialized = bound.deserialize({})
        self.assertEqual(deserialized["querystring"], {"foo": "bar"})
        self.assertEqual(deserialized["body"], {"foo": "beer"})


class PluralQuerySchemaTest(unittest.TestCase):
    def setUp(self):
        self.schema = schema.PluralQuerySchema()
        self.querystring = {
            "_limit": "2",
            "_sort": "toto,tata",
            "_token": "abc",
            "_since": "1234",
            "_to": "7890",
            "_before": "4567",
            "id": "toot",
            "last_modified": "9874",
            "in_age": "15,16,17",
            "exclude_id": "45aea306-cf9c-44d3-fbe0e5bb3ef6,2400fc9b-1db6-473d-818b68431545",
            "contains_comma_int": "3,4",
            "contains_any_comma_strings": "red,blue",
            "contains_any_object": '{"id": 1234}',
            "contains_user.name": '["Ethan","Remy"]',
            "contains_any_user.height": "[165, 200]",
            "contains_aliases": '[{"ls": "ll -a"}, {"rimraf": "rm -rf"}]',
        }

    def test_decode_valid_querystring(self):
        self.maxDiff = None
        deserialized = self.schema.deserialize(self.querystring)
        self.assertEqual(
            deserialized,
            {
                "_limit": 2,
                "_sort": ["toto", "tata"],
                "_token": "abc",
                "_since": 1234,
                "_to": 7890,
                "_before": 4567,
                "id": "toot",
                "last_modified": 9874,
                "contains_comma_int": ["3,4"],
                "contains_any_comma_strings": ["red,blue"],
                "contains_any_object": [{"id": 1234}],
                "contains_any_user.height": [165, 200],
                "contains_user.name": ["Ethan", "Remy"],
                "contains_aliases": [{"ls": "ll -a"}, {"rimraf": "rm -rf"}],
                "in_age": [15, 16, 17],
                "exclude_id": [
                    "45aea306-cf9c-44d3-fbe0e5bb3ef6",
                    "2400fc9b-1db6-473d-818b68431545",
                ],
            },
        )

    def test_raises_invalid_for_to_big_integer_in_limit(self):
        querystring = self.querystring.copy()
        querystring["_limit"] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_to_big_integer_in_since(self):
        querystring = self.querystring.copy()
        querystring["_since"] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_to_big_integer_in_to(self):
        querystring = self.querystring.copy()
        querystring["_to"] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_to_big_integer_in_before(self):
        querystring = self.querystring.copy()
        querystring["_before"] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_to_big_integer_in_last_modified(self):
        querystring = self.querystring.copy()
        querystring["last_modified"] = schema.POSTGRESQL_MAX_INTEGER_VALUE + 1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_limit(self):
        querystring = self.querystring.copy()
        querystring["_limit"] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_since(self):
        querystring = self.querystring.copy()
        querystring["_since"] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_to(self):
        querystring = self.querystring.copy()
        querystring["_to"] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_before(self):
        querystring = self.querystring.copy()
        querystring["_before"] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)

    def test_raises_invalid_for_negative_integer_in_last_modified(self):
        querystring = self.querystring.copy()
        querystring["last_modified"] = -1
        self.assertRaises(colander.Invalid, self.schema.deserialize, querystring)


class ResourceReponsesTest(unittest.TestCase):
    def setUp(self):
        self.handler = schema.ResourceReponses()
        self.resource = colander.MappingSchema(title="fake")
        self.obj = schema.ObjectSchema().bind(data=self.resource)

    def test_get_and_bind_assign_resource_schema_to_object_endpoint(self):
        responses = self.handler.get_and_bind("object", "get", object=self.obj)
        ok_response = responses["200"]
        self.assertEqual(self.obj["data"], ok_response["body"]["data"])

    def test_get_and_bind_assign_resource_schema_to_plural_endpoint(self):
        responses = self.handler.get_and_bind("plural", "get", object=self.obj)
        ok_response = responses["200"]
        # XXX: Data is repeated because it's a colander sequence type index
        self.assertEqual(self.obj["data"], ok_response["body"]["data"]["data"])

    def test_responses_doesnt_have_permissions_if_not_bound(self):
        responses = self.handler.get_and_bind("object", "get", object=self.obj)
        ok_response = responses["200"]
        self.assertNotIn("permissions", ok_response["body"])



import unittest
from unittest import mock
from uuid import UUID

from pyramid.httpexceptions import HTTPBadRequest

from kinto.core.errors import ERRORS, http_error
from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.testing import FormattedErrorMixin, get_user_headers
from kinto.core.utils import hmac_digest

from ..support import MINIMALIST_RECORD, BaseWebTest


class DefaultBucketWebTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["includes"] = "kinto.plugins.default_bucket"
        return settings


class DefaultBucketViewTest(FormattedErrorMixin, DefaultBucketWebTest):
    bucket_url = "/buckets/default"
    collection_url = "/buckets/default/collections/tasks"

    def test_default_bucket_exists_and_has_user_id(self):
        bucket = self.app.get(self.bucket_url, headers=self.headers)
        result = bucket.json
        settings = self.app.app.registry.settings
        default_bucket_hmac_secret = settings["default_bucket_hmac_secret"]
        bucket_id = hmac_digest(default_bucket_hmac_secret, self.principal)[:32]

        self.assertEqual(result["data"]["id"], str(UUID(bucket_id)))
        self.assertEqual(result["permissions"]["write"], [self.principal])

    def test_default_bucket_exists_and_has_user_id_fallback_to_hmac_secret(self):
        bucket = self.app.get(self.bucket_url, headers=self.headers)
        result = bucket.json
        settings = self.app.app.registry.settings
        hmac_secret = settings["userid_hmac_secret"]
        bucket_id = hmac_digest(hmac_secret, self.principal)[:32]

        self.assertEqual(result["data"]["id"], str(UUID(bucket_id)))
        self.assertEqual(result["permissions"]["write"], [self.principal])

    def test_default_bucket_hmac_secret_define(self):
        settings = self.get_app_settings(
            {"default_bucket_hmac_secret": "bucket_id_salt", "userid_hmac_secret": "secret"}
        )
        secret = settings.get("default_bucket_hmac_secret", settings["userid_hmac_secret"])
        value = hmac_digest(secret, "input data")

        self.assertEqual(secret, "bucket_id_salt")
        self.assertTrue(value.startswith("70f09c2d07853887d6b72bc70517f96377"))

    def test_default_bucket_hmac_secret_not_define_fallback_to_userid_hmac_secret(self):
        settings = self.get_app_settings({"userid_hmac_secret": "secret"})
        if "default_bucket_hmac_secret" in settings:
            del settings["default_bucket_hmac_secret"]
        secret = settings.get("default_bucket_hmac_secret", settings["userid_hmac_secret"])
        value = hmac_digest(secret, "input data")

        self.assertEqual(secret, "secret")
        self.assertTrue(value.startswith("6b9fc51c8e13506e9aee1f4c4bdbb650cd"))

    def test_default_bucket_can_still_be_explicitly_created(self):
        bucket = {"permissions": {"read": ["system.Everyone"]}}
        resp = self.app.put_json(self.bucket_url, bucket, headers=self.headers)
        result = resp.json
        self.assertIn("system.Everyone", result["permissions"]["read"])

    def test_default_bucket_collection_can_still_be_explicitly_created(self):
        collection = {"data": {"synced": True}}
        resp = self.app.put_json(self.collection_url, collection, headers=self.headers)
        result = resp.json
        self.assertIn("synced", result["data"])
        self.assertTrue(result["data"]["synced"])
        resp = self.app.get(self.collection_url, headers=self.headers)
        result = resp.json
        self.assertIn("synced", result["data"])
        self.assertTrue("synced", result["data"]["synced"])

    def test_default_bucket_can_be_created_with_simple_put(self):
        self.app.put(self.bucket_url, headers=get_user_headers("bob"), status=201)

    def test_default_bucket_collections_are_automatically_created(self):
        self.app.get(self.collection_url, headers=self.headers, status=200)

    def test_adding_a_task_for_bob_doesnt_add_it_for_alice(self):
        record = {**MINIMALIST_RECORD}
        resp = self.app.post_json(
            self.collection_url + "/records", record, headers=get_user_headers("bob")
        )
        record_id = "{}/records/{}".format(self.collection_url, resp.json["data"]["id"])
        resp = self.app.get(record_id, headers=get_user_headers("alice"), status=404)

    def test_unauthenticated_bucket_access_raises_json_401(self):
        resp = self.app.get(self.bucket_url, status=401)
        self.assertEqual(
            resp.json["message"], "Please authenticate yourself to use this endpoint."
        )

    def test_bucket_id_is_an_uuid_with_dashes(self):
        bucket = self.app.get(self.bucket_url, headers=self.headers)
        bucket_id = bucket.json["data"]["id"]
        self.assertIn("-", bucket_id)
        try:
            UUID(bucket_id)
        except ValueError:
            self.fail("bucket_id: {} is not a valid UUID.".format(bucket_id))

    def test_second_call_on_default_bucket_doesnt_raise_a_412(self):
        self.app.get(self.bucket_url, headers=self.headers)
        self.app.get(self.bucket_url, headers=self.headers)

    def test_second_call_on_default_bucket_collection_doesnt_raise_a_412(self):
        self.app.get(self.collection_url, headers=self.headers)
        self.app.get(self.collection_url, headers=self.headers)

    def test_querystring_parameters_are_taken_into_account(self):
        self.app.get(
            self.collection_url + "/records?_since=invalid", headers=self.headers, status=400
        )

    def test_option_is_possible_without_authentication_for_default(self):
        headers = "authorization,content-type"
        self.app.options(
            self.collection_url + "/records",
            headers={
                "Origin": "http://localhost:8000",
                "Access-Control-Request-Method": "GET",
                "Access-Control-Request-Headers": headers,
            },
        )

    def test_cors_headers_are_provided_on_errors(self):
        resp = self.app.post_json(
            self.collection_url + "/records", MINIMALIST_RECORD, headers=self.headers
        )
        current = resp.json["data"]["last_modified"]
        headers = {
            **self.headers,
            "Origin": "http://localhost:8000",
            "If-None-Match": ('"{}"'.format(current)).encode("utf-8"),
        }
        resp = self.app.get(self.collection_url + "/records", headers=headers, status=304)
        self.assertIn("Access-Control-Allow-Origin", resp.headers)
        self.assertIn("ETag", resp.headers["Access-Control-Expose-Headers"])

    def test_etag_is_present_and_exposed_in_304_error(self):
        resp = self.app.post_json(
            self.collection_url + "/records", MINIMALIST_RECORD, headers=self.headers
        )
        current = resp.json["data"]["last_modified"]
        headers = {
            **self.headers,
            "Origin": "http://localhost:8000",
            "If-None-Match": ('"{}"'.format(current)).encode("utf-8"),
        }
        resp = self.app.get(self.collection_url + "/records", headers=headers, status=304)
        self.assertIn("Access-Control-Expose-Headers", resp.headers)
        self.assertIn("ETag", resp.headers)
        self.assertIn("ETag", resp.headers["Access-Control-Expose-Headers"])

    def test_bucket_id_starting_with_default_can_still_be_created(self):
        # We need to create the bucket first since it is not the default bucket
        resp = self.app.put(
            self.bucket_url.replace("default", "default-1234"), headers=self.headers, status=201
        )
        bucket_id = resp.json["data"]["id"]
        self.assertEqual(bucket_id, "default-1234")

        # We can then create the collection
        collection_url = "/buckets/default-1234/collections/default"
        self.app.put(collection_url, headers=self.headers, status=201)
        resp = self.app.get("/buckets/default-1234/collections", headers=self.headers)
        self.assertEqual(resp.json["data"][0]["id"], "default")

    def test_bucket_id_in_body_can_be_default(self):
        self.app.put_json(
            "/buckets/default", {"data": {"id": "default"}}, headers=self.headers, status=201
        )

    def test_default_bucket_objects_are_checked_only_once_in_batch(self):
        batch = {"requests": []}
        nb_create = 25
        for i in range(nb_create):
            request = {
                "method": "POST",
                "path": self.collection_url + "/records",
                "body": MINIMALIST_RECORD,
            }
            batch["requests"].append(request)

        with mock.patch.object(self.storage, "create", wraps=self.storage.create) as patched:
            self.app.post_json("/batch", batch, headers=self.headers)
            self.assertEqual(patched.call_count, nb_create + 2)

    def test_parent_collection_is_taken_from_the_one_created_in_batch(self):
        batch = {"requests": []}
        nb_create = 25
        for i in range(nb_create):
            request = {
                "method": "POST",
                "path": self.collection_url + "/records",
                "body": MINIMALIST_RECORD,
            }
            batch["requests"].append(request)

        with mock.patch.object(self.storage, "create", wraps=self.storage.create) as patched:
            self.app.post_json("/batch", batch, headers=self.headers)
            # Called twice only: bucket + collection ids unicity.
            self.assertEqual(patched.call_count, 25 + 2)

    def test_parent_collection_is_taken_from_the_one_checked_in_batch(self):
        # Create it first.
        self.app.put(self.collection_url, headers=self.headers, status=201)

        batch = {"requests": []}
        nb_create = 25
        for i in range(nb_create):
            request = {
                "method": "POST",
                "path": self.collection_url + "/records",
                "body": MINIMALIST_RECORD,
            }
            batch["requests"].append(request)

        with mock.patch.object(self.storage, "create", wraps=self.storage.create) as patched:
            self.app.post_json("/batch", batch, headers=self.headers)
            # Called twice only: bucket + collection ids unicity.
            self.assertEqual(patched.call_count, 25 + 2)

    def test_collection_id_is_validated(self):
        collection_url = "/buckets/default/collections/__files__/records"
        self.app.get(collection_url, headers=self.headers, status=400)

    def test_collection_id_does_not_support_unicode(self):
        collection_url = "/buckets/default/collections/%E8%A6%8B/records"
        self.app.get(collection_url, headers=self.headers, status=400)

    def test_405_is_a_valid_formatted_error(self):
        response = self.app.post(self.collection_url, headers=self.headers, status=405)
        self.assertFormattedError(
            response,
            405,
            ERRORS.METHOD_NOT_ALLOWED,
            "Method Not Allowed",
            "Method not allowed on this endpoint.",
        )

    def test_formatted_error_are_passed_through(self):
        # Create the parent objects
        self.app.post(self.collection_url + "/records", headers=self.headers)

        # Simulate a validation error.
        fake400 = http_error(HTTPBadRequest(), errno=ERRORS.INVALID_PARAMETERS, message="Yop")
        with mock.patch.object(self.storage, "create") as mocked:
            mocked.side_effect = [
                {"id": "abc", "last_modified": 43},
                {"id": "abc", "last_modified": 44},
                fake400,
            ]
            resp = self.app.post(
                self.collection_url + "/records", headers=self.headers, status=400
            )
            self.assertEqual(resp.body, fake400.body)

    def test_trailing_slash_redirection_works_for_default_bucket(self):
        collection_url = "/buckets/default/"
        resp = self.app.get(collection_url, headers=self.headers, status=307)
        assert resp.headers["Location"].endswith("/buckets/default")

    def test_trailing_slash_redirection_works_for_collections(self):
        collection_url = "/buckets/default/collections/"
        resp = self.app.get(collection_url, headers=self.headers, status=307)
        assert resp.headers["Location"].endswith("/buckets/default/collections")

    def test_trailing_slash_redirection_works_for_collection(self):
        collection_url = "/buckets/default/collections/foo/"
        resp = self.app.get(collection_url, headers=self.headers, status=307)
        assert resp.headers["Location"].endswith("/buckets/default/collections/foo")

    def test_trailing_slash_redirection_works_for_records(self):
        records_url = "/buckets/default/collections/foo/records/"
        resp = self.app.get(records_url, headers=self.headers, status=307)
        assert resp.headers["Location"].endswith("/buckets/default/collections/foo/records")

    def test_trailing_slash_redirection_works_for_record(self):
        records_url = "/buckets/default/collections/foo/records/bar/"
        resp = self.app.get(records_url, headers=self.headers, status=307)
        assert resp.headers["Location"].endswith("/buckets/default/collections/foo/records/bar")


class HelloViewTest(DefaultBucketWebTest):
    def test_returns_bucket_id_and_url_if_authenticated(self):
        response = self.app.get("/", headers=self.headers)
        userinfo = response.json["user"]
        self.assertIn("id", userinfo)
        self.assertIn("principals", userinfo)
        self.assertIn("bucket", userinfo)
        self.assertEqual(userinfo["bucket"], "ddaf8694-fa9e-2949-ed0e-77198a7907fb")

    def test_flush_capability_if_enabled(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("default_bucket", capabilities)


_events = []


def load_from_config(config, prefix):
    def listener(event):
        _events.append(event)

    return listener


class EventsTest(DefaultBucketWebTest):
    def tearDown(self):
        super().tearDown()
        del _events[:]

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings = {
            **settings,
            "event_listeners": "testevent",
            "event_listeners.testevent.use": "tests.plugins.test_default_bucket",
        }
        return settings

    def test_an_event_is_sent_on_implicit_bucket_creation(self):
        bucket_url = "/buckets/default"
        self.app.get(bucket_url, headers=self.headers)
        assert len(_events) == 1
        assert "last_modified" in _events[-1].impacted_objects[0]["new"]
        payload = _events[-1].payload
        assert payload["resource_name"] == "bucket"
        assert payload["action"] == "create"

    def test_an_event_is_sent_on_implicit_collection_creation(self):
        collection_url = "/buckets/default/collections/articles"
        self.app.get(collection_url, headers=self.headers)
        assert len(_events) == 2
        payload = _events[-1].payload
        assert payload["resource_name"] == "collection"
        assert payload["action"] == "create"

    def test_events_sent_on_bucket_and_collection_creation(self):
        records_uri = "/buckets/default/collections/articles/records/implicit"
        body = {"data": {"implicit": "creations"}}
        self.app.put_json(records_uri, body, headers=self.headers)

        assert len(_events) == 3

        # Implicit creation of bucket
        resp = self.app.get("/", headers=self.headers)
        bucket_id = resp.json["user"]["bucket"]
        assert "subpath" not in _events[0].payload
        assert _events[0].payload["action"] == "create"
        assert _events[0].payload["bucket_id"] == bucket_id
        assert _events[0].payload["uri"] == "/buckets/{}".format(bucket_id)

        # Implicit creation of collection
        assert "subpath" not in _events[1].payload
        assert _events[1].payload["action"] == "create"
        assert _events[1].payload["resource_name"] == "collection"
        assert _events[1].payload["bucket_id"] == bucket_id
        assert _events[1].payload["collection_id"] == "articles"
        assert _events[1].payload["uri"] == "/buckets/{}/collections/articles".format(bucket_id)

        # Creation of record
        assert _events[2].payload["action"] == "create"
        assert _events[2].payload["resource_name"] == "record"
        assert _events[2].payload["bucket_id"] == bucket_id
        assert _events[2].payload["collection_id"] == "articles"
        assert _events[2].payload["uri"] == records_uri.replace("default", bucket_id)

    def test_second_call_on_default_bucket_doesnt_send_create_events(self):
        bucket_url = "/buckets/default"
        self.app.get(bucket_url, headers=self.headers)
        # An event is generated -- it's the same event from
        # test_an_event_is_sent_on_implicit_bucket_creation.
        assert len(_events) == 1
        self.app.get(bucket_url, headers=self.headers)
        # Bucket shouldn't be created again, so there shouldn't be any
        # more events.
        assert len(_events) == 1

    def test_second_call_on_default_bucket_collection_doesnt_send_create_events(self):
        collection_url = "/buckets/default/collections/articles"
        self.app.get(collection_url, headers=self.headers)
        # Events are generated -- they're the same event from
        # test_an_event_is_sent_on_implicit_collection_creation.
        assert len(_events) == 2
        self.app.get(collection_url, headers=self.headers)
        # Neither the bucket nor the collection should be created
        # again, so there shouldn't be any more events.
        assert len(_events) == 2


class ReadonlyDefaultBucket(DefaultBucketWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["readonly"] = True
        return settings

    def test_implicit_creation_is_rejected(self):
        self.app.get("/buckets/default", headers=self.headers, status=405)


class BackendErrorTest(DefaultBucketWebTest):
    def setUp(self):
        super().setUp()
        self.patcher = mock.patch.object(
            self.storage, "create", side_effect=storage_exceptions.BackendError()
        )
        self.addCleanup(self.patcher.stop)

    def test_implicit_bucket_creation_raises_503_if_backend_fails(self):
        self.patcher.start()
        self.app.get("/buckets/default", headers=self.headers, status=503)

    def test_implicit_collection_creation_raises_503_if_backend_fails(self):
        self.app.get("/buckets/default", headers=self.headers)
        self.patcher.start()
        self.app.get("/buckets/default/collections/articles", headers=self.headers, status=503)



import re
import uuid

from pyramid import httpexceptions
from pyramid.events import subscriber

from kinto.core import Service
from kinto.core.errors import http_error, raise_invalid
from kinto.core.events import ACTIONS, ResourceChanged
from kinto.core.storage import exceptions as storage_exceptions

from ..mails import Emailer
from ..utils import (
    cache_reset_password,
    delete_cached_validation_key,
    get_cached_validation_key,
    hash_password,
)
from . import DEFAULT_EMAIL_REGEXP


# Account validation (enable in the settings).
validation = Service(
    name="account-validation",
    path="/accounts/{user_id}/validate/{activation_key}",
    description="Validate an account",
)


def check_validation_key(activation_key, username, registry):
    """Given a username, compare the activation-key provided with the one from the cache."""
    cache_result = get_cached_validation_key(username, registry)

    if cache_result == activation_key:
        delete_cached_validation_key(username, registry)  # We're done with the activation key.
        return True
    return False


@validation.post()
def post_validation(request):
    user_id = request.matchdict["user_id"]
    activation_key = request.matchdict["activation_key"]

    parent_id = user_id
    try:
        user = request.registry.storage.get(
            parent_id=parent_id, resource_name="account", object_id=user_id
        )
    except storage_exceptions.ObjectNotFoundError:
        # Don't give information on the existence of a user id: return a generic error message.
        error_details = {"message": "Account ID and activation key do not match"}
        raise http_error(httpexceptions.HTTPForbidden(), **error_details)

    if not check_validation_key(activation_key, user_id, request.registry):
        error_details = {"message": "Account ID and activation key do not match"}
        raise http_error(httpexceptions.HTTPForbidden(), **error_details)

    # User is now validated.
    new = user.copy()
    new["validated"] = True

    result = request.registry.storage.update(
        parent_id=parent_id, resource_name="account", object_id=user_id, obj=new
    )
    request.notify_resource_event(
        parent_id=parent_id,
        timestamp=result["last_modified"],
        data=result,
        action=ACTIONS.UPDATE,
        old=user,
        resource_name="account",
    )
    return new


# Password reset.
reset_password = Service(
    name="reset-password",
    path="/accounts/{user_id}/reset-password",
    description="Send a temporary reset password by mail for an account",
)


@reset_password.post()
def post_reset_password(request):
    user_id = request.matchdict["user_id"]

    parent_id = user_id
    try:
        user = request.registry.storage.get(
            parent_id=parent_id, resource_name="account", object_id=user_id
        )
    except storage_exceptions.ObjectNotFoundError:
        # Don't give information on the existence of a user id: return a generic message.
        return {"message": "A temporary reset password has been sent by mail"}

    settings = request.registry.settings

    user_email = user["id"]
    email_regexp = settings.get("account_validation.email_regexp", DEFAULT_EMAIL_REGEXP)
    compiled_email_regexp = re.compile(email_regexp)
    if not compiled_email_regexp.match(user_email):
        error_details = {
            "name": "data.id",
            "description": f"The user id should match {email_regexp}.",
        }
        raise_invalid(request, **error_details)

    reset_password = str(uuid.uuid4())
    hashed_reset_password = hash_password(reset_password)
    cache_reset_password(hashed_reset_password, user_id, request.registry)

    # Send a temporary reset password by mail.
    Emailer(request, user).send_temporary_reset_password(reset_password)

    return {"message": "A temporary reset password has been sent by mail"}


# Send confirmation email on account activation if account validation is enabled.
@subscriber(ResourceChanged, for_resources=("account",), for_actions=(ACTIONS.UPDATE,))
def on_account_activated(event):
    request = event.request
    settings = request.registry.settings
    if not settings.get("account_validation", False):
        return

    for impacted_object in event.impacted_objects:
        old_account = impacted_object["old"]
        account = impacted_object["new"]
        if old_account.get("validated", True) or not account.get("validated", False):
            # It's not an account activation, bail.
            continue

        # Send a confirmation email.
        Emailer(request, account).send_confirmation()



import copy

from pyramid.httpexceptions import HTTPInsufficientStorage

from kinto.core.errors import ERRORS, http_error
from kinto.core.storage.exceptions import ObjectNotFoundError
from kinto.core.utils import instance_uri

from .utils import record_size


QUOTA_RESOURCE_NAME = "quota"
BUCKET_QUOTA_OBJECT_ID = "bucket_info"
COLLECTION_QUOTA_OBJECT_ID = "collection_info"


def raise_insufficient_storage(message):
    raise http_error(HTTPInsufficientStorage(), errno=ERRORS.FORBIDDEN.value, message=message)


def get_bucket_settings(settings, bucket_id, name):
    return settings.get(
        # Bucket specific
        f"quotas.bucket_{bucket_id}_{name}",
        # Global to all buckets
        settings.get(f"quotas.bucket_{name}", None),
    )


def get_collection_settings(settings, bucket_id, collection_id, name):
    return settings.get(
        # Specific for a given bucket collection
        f"quotas.collection_{bucket_id}_{collection_id}_{name}",
        # Specific to given bucket collections
        settings.get(
            f"quotas.collection_{bucket_id}_{name}",
            # Global to all buckets collections
            settings.get(f"quotas.collection_{name}", None),
        ),
    )


def on_resource_changed(event):
    """
    Everytime an object is created/changed/deleted, we update the
    bucket counters.

    If a new object exceeds the quotas, we reject the request.
    """
    payload = event.payload
    action = payload["action"]
    resource_name = payload["resource_name"]

    if action == "delete" and resource_name == "bucket":
        # Deleting a bucket already deletes everything underneath (including
        # quotas info). See kinto/views/bucket.
        return

    settings = event.request.registry.settings

    event_uri = payload["uri"]
    bucket_id = payload["bucket_id"]
    bucket_uri = instance_uri(event.request, "bucket", id=bucket_id)
    collection_id = None
    collection_uri = None
    if "collection_id" in payload:
        collection_id = payload["collection_id"]
        collection_uri = instance_uri(
            event.request, "collection", bucket_id=bucket_id, id=collection_id
        )

    bucket_max_bytes = get_bucket_settings(settings, bucket_id, "max_bytes")
    bucket_max_items = get_bucket_settings(settings, bucket_id, "max_items")
    bucket_max_bytes_per_item = get_bucket_settings(settings, bucket_id, "max_bytes_per_item")
    collection_max_bytes = get_collection_settings(settings, bucket_id, collection_id, "max_bytes")
    collection_max_items = get_collection_settings(settings, bucket_id, collection_id, "max_items")
    collection_max_bytes_per_item = get_collection_settings(
        settings, bucket_id, collection_id, "max_bytes_per_item"
    )

    max_bytes_per_item = collection_max_bytes_per_item or bucket_max_bytes_per_item

    storage = event.request.registry.storage

    targets = []
    for impacted in event.impacted_objects:
        target = impacted["new" if action != "delete" else "old"]
        # On POST .../records, the URI does not contain the newly created
        # record id.
        obj_id = target["id"]
        parts = event_uri.split("/")
        if resource_name in parts[-1]:
            parts.append(obj_id)
        else:
            # Make sure the id is correct on grouped events.
            parts[-1] = obj_id
        uri = "/".join(parts)

        old = impacted.get("old", {})
        new = impacted.get("new", {})

        targets.append((uri, obj_id, old, new))

    try:
        bucket_info = copy.deepcopy(
            storage.get(
                parent_id=bucket_uri,
                resource_name=QUOTA_RESOURCE_NAME,
                object_id=BUCKET_QUOTA_OBJECT_ID,
            )
        )
    except ObjectNotFoundError:
        bucket_info = {"collection_count": 0, "record_count": 0, "storage_size": 0}

    collection_info = {"record_count": 0, "storage_size": 0}
    if collection_id:
        try:
            collection_info = copy.deepcopy(
                storage.get(
                    parent_id=collection_uri,
                    resource_name=QUOTA_RESOURCE_NAME,
                    object_id=COLLECTION_QUOTA_OBJECT_ID,
                )
            )
        except ObjectNotFoundError:
            pass

    # Update the bucket quotas values for each impacted record.
    for uri, obj_id, old, new in targets:
        old_size = record_size(old)
        new_size = record_size(new)

        if max_bytes_per_item is not None and action != "delete":
            if new_size > max_bytes_per_item:
                message = f'Maximum bytes per object exceeded " "({new_size} > {max_bytes_per_item} Bytes.'
                raise_insufficient_storage(message)

        if action == "create":
            bucket_info["storage_size"] += new_size
            if resource_name == "collection":
                bucket_info["collection_count"] += 1
                collection_info["storage_size"] += new_size
            if resource_name == "record":
                bucket_info["record_count"] += 1
                collection_info["record_count"] += 1
                collection_info["storage_size"] += new_size
        elif action == "update":
            bucket_info["storage_size"] -= old_size
            bucket_info["storage_size"] += new_size
            if resource_name in ("collection", "record"):
                collection_info["storage_size"] -= old_size
                collection_info["storage_size"] += new_size
        else:  # action == 'delete':
            bucket_info["storage_size"] -= old_size
            if resource_name == "collection":
                collection_uri = uri
                bucket_info["collection_count"] -= 1
                # When we delete the collection all the records in it
                # are deleted without notification.
                collection_records = storage.list_all(
                    resource_name="record", parent_id=collection_uri
                )
                for r in collection_records:
                    old_record_size = record_size(r)
                    bucket_info["record_count"] -= 1
                    bucket_info["storage_size"] -= old_record_size
                    collection_info["record_count"] -= 1
                    collection_info["storage_size"] -= old_record_size
                collection_info["storage_size"] -= old_size

            if resource_name == "record":
                bucket_info["record_count"] -= 1
                collection_info["record_count"] -= 1
                collection_info["storage_size"] -= old_size

    if bucket_max_bytes is not None:
        if bucket_info["storage_size"] > bucket_max_bytes:
            message = (
                "Bucket maximum total size exceeded "
                f"({bucket_info['storage_size']} > {bucket_max_bytes} Bytes). "
            )
            raise_insufficient_storage(message)

    if bucket_max_items is not None:
        if bucket_info["record_count"] > bucket_max_items:
            message = (
                "Bucket maximum number of objects exceeded "
                f"({bucket_info['record_count']} > {bucket_max_items} objects)."
            )
            raise_insufficient_storage(message)

    if collection_max_bytes is not None:
        if collection_info["storage_size"] > collection_max_bytes:
            message = (
                "Collection maximum size exceeded "
                f"({collection_info['storage_size']} > {collection_max_bytes} Bytes)."
            )
            raise_insufficient_storage(message)

    if collection_max_items is not None:
        if collection_info["record_count"] > collection_max_items:
            message = (
                "Collection maximum number of objects exceeded "
                f"({collection_info['record_count']} > {collection_max_items} objects)."
            )
            raise_insufficient_storage(message)

    storage.update(
        parent_id=bucket_uri,
        resource_name=QUOTA_RESOURCE_NAME,
        object_id=BUCKET_QUOTA_OBJECT_ID,
        obj=bucket_info,
    )

    if collection_id:
        if action == "delete" and resource_name == "collection":
            # Deleting a collection already deletes everything underneath
            # (including quotas info). See kinto/views/collection.
            return
        else:
            storage.update(
                parent_id=collection_uri,
                resource_name=QUOTA_RESOURCE_NAME,
                object_id=COLLECTION_QUOTA_OBJECT_ID,
                obj=collection_info,
            )



from kinto.authorization import _inherited_permissions, _relative_object_uri, _resource_endpoint
from kinto.core.testing import unittest


class ResourceEndpointTest(unittest.TestCase):
    def test_resource_endpoint_return_right_type_for_key(self):
        self.assertEqual(
            _resource_endpoint("/buckets/bid/collections/cid/records/rid"), ("record", False)
        )
        self.assertEqual(_resource_endpoint("/buckets/bid/collections/cid"), ("collection", False))
        self.assertEqual(_resource_endpoint("/buckets"), ("", False))
        self.assertEqual(_resource_endpoint("/buckets/bid"), ("bucket", False))
        self.assertEqual(_resource_endpoint("/buckets/bid/history/xx"), ("history", False))
        self.assertEqual(_resource_endpoint("/buckets/bid/groups/moderators"), ("group", False))

    def test_resource_endpoint_return_right_type_for_children_collection(self):
        self.assertEqual(
            _resource_endpoint("/buckets/bid/collections/cid/records"), ("collection", True)
        )
        self.assertEqual(_resource_endpoint("/buckets/bid/collections"), ("bucket", True))
        self.assertEqual(_resource_endpoint("/buckets/bid/history"), ("bucket", True))
        self.assertEqual(_resource_endpoint("/buckets/bid/groups"), ("bucket", True))


class RelativeObjectUri(unittest.TestCase):
    record_uri = "/buckets/blog/collections/articles/records/article1"
    records_uri = "/buckets/blog/collections/articles/records"
    collection_uri = "/buckets/blog/collections/articles"
    collections_uri = "/buckets/blog/collections"
    group_uri = "/buckets/blog/groups/moderators"
    groups_uri = "/buckets/blog/groups"
    bucket_uri = "/buckets/blog"
    buckets_uri = "/buckets"

    def test_related_object_uri_can_construct_parents_set_uris(self):
        record_uri = "/buckets/bid/collections/cid/records/rid"
        self.assertEqual(_relative_object_uri("record", record_uri), record_uri)

        self.assertEqual(_relative_object_uri("collection", self.record_uri), self.collection_uri)
        self.assertEqual(_relative_object_uri("collection", self.records_uri), self.collection_uri)

        # Can build bucket_uri from obj_parts
        self.assertEqual(_relative_object_uri("bucket", self.record_uri), self.bucket_uri)

        # Can build group_uri from group obj_parts
        self.assertEqual(_relative_object_uri("group", self.group_uri), self.group_uri)

        # Can build bucket_uri from group obj_parts
        self.assertEqual(_relative_object_uri("bucket", self.group_uri), self.bucket_uri)
        self.assertEqual(_relative_object_uri("bucket", self.groups_uri), self.bucket_uri)

    def test_relative_object_uri_fail_construct_children_set_uris(self):
        # Cannot build record_uri from bucket obj_parts
        self.assertRaises(ValueError, _relative_object_uri, "record", self.bucket_uri)

        # Cannot build collection_uri from obj_parts
        self.assertRaises(ValueError, _relative_object_uri, "collection", self.bucket_uri)

        # Cannot build bucket_uri from empty obj_parts
        self.assertRaises(ValueError, _relative_object_uri, "collection", "")


class PermissionInheritanceTest(unittest.TestCase):
    record_uri = "/buckets/blog/collections/articles/records/article1"
    collection_uri = "/buckets/blog/collections/articles"
    group_uri = "/buckets/blog/groups/moderators"
    bucket_uri = "/buckets/blog"

    def test_relative_object_uri_fail_on_wrong_type(self):
        self.assertRaises(ValueError, _relative_object_uri, "schema", self.record_uri)

    def test_inherited_permissions_supports_buckets_named_collections(self):
        uri = "/buckets/collections"
        self.assertEqual(set(_inherited_permissions(uri, "write")), set([(uri, "write")]))

    def test_inherited_permissions_for_bucket_permission(self):
        # write
        self.assertEqual(
            _inherited_permissions(self.bucket_uri, "write"), [(self.bucket_uri, "write")]
        )
        # read
        self.assertEqual(
            set(_inherited_permissions(self.bucket_uri, "read")),
            set(
                [
                    (self.bucket_uri, "write"),
                    (self.bucket_uri, "read"),
                    (self.bucket_uri, "collection:create"),
                    (self.bucket_uri, "group:create"),
                ]
            ),
        )

        # group:create
        groups_uri = self.bucket_uri + "/groups"
        self.assertEqual(
            set(_inherited_permissions(groups_uri, "group:create")),
            set([(self.bucket_uri, "write"), (self.bucket_uri, "group:create")]),
        )

        # collection:create
        collections_uri = self.bucket_uri + "/collections"
        self.assertEqual(
            set(_inherited_permissions(collections_uri, "collection:create")),
            set([(self.bucket_uri, "write"), (self.bucket_uri, "collection:create")]),
        )

    def test_inherited_permissions_for_group_permission(self):
        # write
        self.assertEqual(
            _inherited_permissions(self.group_uri, "write"),
            [(self.group_uri, "write"), (self.bucket_uri, "write")],
        )
        # read
        self.assertEqual(
            set(_inherited_permissions(self.group_uri, "read")),
            set(
                [
                    (self.bucket_uri, "write"),
                    (self.bucket_uri, "read"),
                    (self.group_uri, "write"),
                    (self.group_uri, "read"),
                ]
            ),
        )

    def test_inherited_permissions_for_collection_permission(self):
        # write
        self.assertEqual(
            _inherited_permissions(self.collection_uri, "write"),
            [(self.collection_uri, "write"), (self.bucket_uri, "write")],
        )
        # read
        self.assertEqual(
            set(_inherited_permissions(self.collection_uri, "read")),
            set(
                [
                    (self.bucket_uri, "write"),
                    (self.bucket_uri, "read"),
                    (self.collection_uri, "write"),
                    (self.collection_uri, "read"),
                    (self.collection_uri, "record:create"),
                ]
            ),
        )
        # records:create
        records_uri = self.collection_uri + "/records"
        self.assertEqual(
            set(_inherited_permissions(records_uri, "record:create")),
            set(
                [
                    (self.bucket_uri, "write"),
                    (self.collection_uri, "write"),
                    (self.collection_uri, "record:create"),
                ]
            ),
        )

    def test_inherited_permissions_for_record_permission(self):
        # write
        self.assertEqual(
            set(_inherited_permissions(self.record_uri, "write")),
            set(
                [
                    (self.bucket_uri, "write"),
                    (self.collection_uri, "write"),
                    (self.record_uri, "write"),
                ]
            ),
        )
        # read
        self.assertEqual(
            set(_inherited_permissions(self.record_uri, "read")),
            set(
                [
                    (self.bucket_uri, "write"),
                    (self.bucket_uri, "read"),
                    (self.collection_uri, "write"),
                    (self.collection_uri, "read"),
                    (self.record_uri, "write"),
                    (self.record_uri, "read"),
                ]
            ),
        )

    def test_inherited_permissions_for_list_of_buckets(self):
        permissions = _inherited_permissions("/buckets", "read")
        self.assertEqual(permissions, [])

    def test_inherited_permissions_for_non_resource_url(self):
        unknown = "/resource/unknown"
        permissions = _inherited_permissions(unknown, "read")
        self.assertEqual(permissions, [])

    def test_inherited_permissions_for_attachment_url(self):
        attachment = "/buckets/bid/collections/cid/records/rid/attachment"
        permissions = _inherited_permissions(attachment, "read")
        self.assertIn(("/buckets/bid/collections/cid/records/rid", "read"), permissions)



import unittest

from pyramid.config import Configurator

from kinto.core import events, utils
from kinto.core.storage import exceptions
from kinto.core.testing import get_user_headers
from kinto.views.admin import get_parent_uri

from .support import MINIMALIST_BUCKET, MINIMALIST_COLLECTION, MINIMALIST_RECORD, BaseWebTest


class DeleteUserDataTest(BaseWebTest, unittest.TestCase):
    """Create some data belonging to a user, then delete the user.

    This user (and all their data) should be deleted. Thus, this user
    (and their data) is "doomed".

    To make sure we don't delete too much data, create some data owned
    by a different user which won't be deleted. This user (and their
    data) will be "safe". We'll use the normal self.headers user (see
    tests.support.BaseWebTest and kinto.core.get_user_headers) for
    this.

    We also want to ensure that the "doomed" user is removed correctly
    from "surviving" permissions and groups, so add the "doomed" user
    to some stuff belonging to the "safe" user too.

    """

    doomed_bucket_url = "/buckets/user_to_delete"
    doomed_collection_url = doomed_bucket_url + "/collections/collection_to_delete"
    doomed_bucket_collection_url = doomed_bucket_url + "/collections/collection"
    safe_bucket_url = "/buckets/user_safe"
    safe_bucket_doomed_collection_url = safe_bucket_url + "/collections/collection_to_delete"
    safe_collection_url = safe_bucket_url + "/collections/collection_safe"
    safe_collection_doomed_record_url = safe_collection_url + "/records/record_to_delete"
    safe_record_url = safe_collection_url + "/records/record_safe"
    safe_group_containing_doomed_user = safe_bucket_url + "/groups/silly-group"
    shared_bucket_url = "/buckets/both"
    doomed_user = get_user_headers("doomed", "trustno1")
    hmac_secret = "secret_hmac_for_userids"
    user_hash = utils.hmac_digest(hmac_secret, "mat:secret")
    user_principal = f"basicauth:{user_hash}"
    doomed_hash = utils.hmac_digest(hmac_secret, "doomed:trustno1")
    doomed_user_principal = f"basicauth:{doomed_hash}"
    delete_user_url = f"/__user_data__/{doomed_user_principal}"
    events = []

    @classmethod
    def listener(cls, event):
        cls.events.append(event)

    @classmethod
    def get_app_settings(cls, extras=None):
        dct = super().get_app_settings(extras=extras)
        dct["userid_hmac_secret"] = cls.hmac_secret
        dct["user-data_delete_principals"] = cls.user_principal
        dct["bucket_create_principals"] = "system.Authenticated"
        return dct

    @classmethod
    def make_app(cls, settings=None, config=None):
        settings = cls.get_app_settings(settings)
        config = Configurator(settings=settings)
        config.add_subscriber(cls.listener, events.ResourceChanged)
        config.commit()
        return super().make_app(settings=settings, config=config)

    def setUp(self):
        del self.events[:]
        super().setUp()
        self.app.put_json(self.doomed_bucket_url, MINIMALIST_BUCKET, headers=self.doomed_user)
        self.app.put_json(
            self.doomed_collection_url, MINIMALIST_COLLECTION, headers=self.doomed_user
        )
        # Temporarily grant write to self.user_principal to allow them
        # to create a collection
        self.app.patch_json(
            self.doomed_bucket_url,
            {"permissions": {"write": [self.user_principal]}},
            headers=self.doomed_user,
        )
        self.app.put_json(
            self.doomed_bucket_collection_url, MINIMALIST_COLLECTION, headers=self.headers
        )
        self.app.patch_json(
            self.doomed_bucket_url, {"permissions": {"write": []}}, headers=self.doomed_user
        )
        self.app.put_json(
            self.safe_bucket_url,
            {**MINIMALIST_BUCKET, "permissions": {"write": [self.doomed_user_principal]}},
            headers=self.headers,
        )
        self.app.put_json(
            self.safe_bucket_doomed_collection_url, MINIMALIST_COLLECTION, headers=self.doomed_user
        )
        self.app.put_json(self.safe_collection_url, MINIMALIST_COLLECTION, headers=self.headers)
        self.app.patch_json(
            self.safe_collection_url,
            {"permissions": {"write": [self.doomed_user_principal]}},
            headers=self.headers,
        )
        self.app.put_json(
            self.safe_collection_doomed_record_url, MINIMALIST_RECORD, headers=self.doomed_user
        )
        self.app.patch_json(
            self.safe_collection_url, {"permissions": {"write": []}}, headers=self.headers
        )
        # This record will automatically have "write" permission for
        # the safe user.
        self.app.put_json(
            self.safe_record_url,
            {**MINIMALIST_RECORD, "permissions": {"read": [self.doomed_user_principal]}},
            headers=self.headers,
        )
        self.app.put_json(
            self.shared_bucket_url,
            {
                **MINIMALIST_BUCKET,
                "permissions": {"write": [self.doomed_user_principal, self.user_principal]},
            },
            headers=self.doomed_user,
        )
        self.app.put_json(
            self.safe_group_containing_doomed_user,
            {"data": {"members": [self.doomed_user_principal, self.user_principal]}},
            headers=self.headers,
        )

        # Actually do the delete. Each `test_` method checks some
        # consequence of what got deleted.
        self.app.delete(self.delete_user_url, headers=self.headers)

    def test_other_methods_are_not_allowed(self):
        for url in ("/__user_data__", self.delete_user_url):
            self.app.head(url, headers=self.headers, status=405)
            self.app.get(url, headers=self.headers, status=405)
            self.app.put(url, headers=self.headers, status=405)
            self.app.patch(url, headers=self.headers, status=405)

    def test_doomed_bucket_was_deleted(self):
        with self.assertRaises(exceptions.ObjectNotFoundError):
            self.storage.get(resource_name="bucket", parent_id="", object_id="user_to_delete")

    def test_doomed_bucket_collection_was_deleted(self):
        # Everything under this bucket should be gone
        with self.assertRaises(exceptions.ObjectNotFoundError):
            self.storage.get(
                resource_name="collection",
                parent_id=self.doomed_bucket_url,
                object_id="collection",
            )

    def test_doomed_bucket_doomed_collection_was_deleted(self):
        with self.assertRaises(exceptions.ObjectNotFoundError):
            self.storage.get(
                resource_name="collection",
                parent_id=self.doomed_bucket_url,
                object_id="collection_to_delete",
            )

    def test_safe_bucket_was_not_deleted(self):
        self.storage.get(resource_name="bucket", parent_id="", object_id="user_safe")

    def test_safe_bucket_doomed_collection_was_deleted(self):
        with self.assertRaises(exceptions.ObjectNotFoundError):
            self.storage.get(
                resource_name="collection",
                parent_id=self.safe_bucket_url,
                object_id="collection_to_delete",
            )

    def test_safe_collection_was_not_deleted(self):
        self.storage.get(
            resource_name="collection", parent_id=self.safe_bucket_url, object_id="collection_safe"
        )

    def test_safe_collection_safe_record_was_not_deleted(self):
        self.storage.get(
            resource_name="record", parent_id=self.safe_collection_url, object_id="record_safe"
        )

    def test_safe_collection_doomed_record_was_deleted(self):
        with self.assertRaises(exceptions.ObjectNotFoundError):
            self.storage.get(
                resource_name="record",
                parent_id=self.safe_collection_url,
                object_id="record_to_delete",
            )

    def test_safe_group_not_deleted(self):
        principals = self.permission.get_user_principals(self.user_principal)
        self.assertEqual(principals, {"/buckets/user_safe/groups/silly-group"})

    def test_safe_group_doomed_user_was_removed(self):
        principals = self.permission.get_user_principals(self.doomed_user_principal)
        self.assertEqual(principals, set())

    def test_doomed_collection_no_longer_has_any_perms(self):
        permissions = self.permission.get_object_permissions(self.doomed_collection_url)
        self.assertEqual(permissions, {})

    def test_doomed_bucket_collection_no_longer_has_any_perms(self):
        permissions = self.permission.get_object_permissions(self.doomed_bucket_collection_url)
        self.assertEqual(permissions, {})

    def test_safe_bucket_no_longer_has_doomed_user_perm(self):
        permissions = self.permission.get_object_permissions(self.safe_bucket_url)
        self.assertEqual(permissions, {"write": {self.user_principal}})

    def test_safe_record_no_longer_has_doomed_user_perm(self):
        permissions = self.permission.get_object_permissions(self.safe_record_url)
        self.assertEqual(permissions, {"write": {self.user_principal}})

    def test_shared_bucket_was_not_deleted(self):
        self.storage.get(resource_name="bucket", parent_id="", object_id="both")

    def test_shared_bucket_permissions_are_correct(self):
        permissions = self.permission.get_object_permissions(self.shared_bucket_url)
        self.assertEqual(permissions, {"write": {self.user_principal}})

    def test_double_delete_is_ok(self):
        self.app.delete(self.delete_user_url, headers=self.headers)

    def test_cannot_delete_user_without_permission(self):
        # Only the user with self.headers can access this URL.
        self.app.delete(self.delete_user_url, headers=self.doomed_user, status=403)

    def test_event_emitted_for_bucket_deletion(self):
        delete_events = [
            e for e in self.events if e.payload["action"] == events.ACTIONS.DELETE.value
        ]
        bucket_events = [e for e in delete_events if e.payload["resource_name"] == "bucket"]
        # There should only be one such event, because all buckets
        # share a parent
        self.assertEqual(len(bucket_events), 1)
        # There should only be one such bucket
        self.assertEqual(len(bucket_events[0].impacted_objects), 1)
        event = bucket_events[0].impacted_objects[0]
        self.assertEqual(event["old"]["id"], "user_to_delete")

    def test_event_not_emitted_for_coalesced_collection_deletion(self):
        # Verify that collection deletion for
        # /buckets/user_to_delete/collections/collection_to_delete
        # was folded into bucket deletion for
        # /buckets/user_to_delete
        delete_events = [
            e for e in self.events if e.payload["action"] == events.ACTIONS.DELETE.value
        ]
        collection_events = [
            e for e in delete_events if e.payload["resource_name"] == "collection"
        ]
        # There should only be one such event, for /buckets/user_safe
        self.assertEqual(len(collection_events), 1)
        self.assertEqual(collection_events[0].payload["bucket_id"], "user_safe")
        # Nothing for user_to_delete, so it must have been coalesced

    def test_event_emitted_for_record_deletion(self):
        delete_events = [
            e for e in self.events if e.payload["action"] == events.ACTIONS.DELETE.value
        ]
        record_events = [e for e in delete_events if e.payload["resource_name"] == "record"]
        # Just one, for safe_collection_doomed_record_url
        self.assertEqual(len(record_events), 1)
        event = record_events[0]
        self.assertEqual(event.payload["bucket_id"], "user_safe")
        self.assertEqual(event.payload["collection_id"], "collection_safe")
        # Just the doomed record, not the safe one
        self.assertEqual(len(event.impacted_objects), 1)
        self.assertEqual(event.impacted_objects[0]["old"]["id"], "record_to_delete")


class GetParentUriTest(unittest.TestCase):
    def test_parent_uri_behaves_sensibly_for_unknown_resources(self):
        unknown_url = "/cities/prague/monuments/castle/foods/mushrooms"
        self.assertEqual(get_parent_uri(unknown_url), "/cities/prague/monuments/castle")

    def test_parent_uri_accepts_pathological_urls(self):
        # This shouldn't ever actually happen, and if it does happen
        # it's probably a sign of a larger incompatibility of this
        # method with the URL scheme
        self.assertEqual(get_parent_uri("/prague"), "")



"""
kinto.core.storage.utils: methods for making it easier to work with kinto storage objects
"""

from kinto.core.storage import Filter
from kinto.core.utils import COMPARISON


BATCH_SIZE = 25


def paginated(storage, *args, sorting, batch_size=BATCH_SIZE, **kwargs):
    """A generator used to access paginated results from storage.list_all.

    :param kwargs: Passed through unchanged to list_all.
    """

    if len(sorting) > 1:
        raise NotImplementedError("FIXME: only supports one-length sorting")  # pragma: nocover
    pagination_direction = COMPARISON.GT if sorting[0].direction > 0 else COMPARISON.LT

    object_pagination = None
    while True:
        objects = storage.list_all(
            sorting=sorting, limit=batch_size, pagination_rules=object_pagination, **kwargs
        )

        if not objects:
            break

        for obj in objects:
            yield obj

        object_pagination = [
            # FIXME: support more than one-length sorting
            [Filter(sorting[0].field, obj[sorting[0].field], pagination_direction)]
        ]



import os

from pyramid.config import Configurator
from pyramid.static import static_view

import kinto.core


here = os.path.dirname(__file__)


def includeme(config):
    config.scan("tests.core.testapp.views")

    # Add an example route with trailing slash (here to serve static files).
    # This is only used to test 404 redirection in ``test_views_errors.py``
    abs_path = os.path.join(here, "static")
    static = static_view(abs_path, use_subpath=True)
    config.add_route("catchall_static", "/static/*subpath")
    config.add_view(static, route_name="catchall_static")


def main(global_config, config=None, **settings):
    global_config.update(settings)
    if config is None:
        config = Configurator(settings=settings)

    kinto.core.initialize(config, version="0.0.1")
    config.include(includeme)
    app = config.make_wsgi_app()
    # Install middleware (no-op if not enabled in setting)
    return kinto.core.install_middlewares(app, settings)



import json
import logging
import unittest
from unittest import mock

from pyramid import testing

from kinto.core import DEFAULT_SETTINGS, JsonLogFormatter, initialization

from .support import BaseWebTest


class RequestSummaryTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        config = testing.setUp()
        config.registry.settings = DEFAULT_SETTINGS
        initialization.setup_logging(config)

        patch = mock.patch("kinto.core.initialization.summary_logger")
        self.mocked = patch.start()
        self.addCleanup(patch.stop)

    def logger_context(self):
        args, kwargs = self.mocked.info.call_args_list[-1]
        return kwargs["extra"]

    def test_standard_info_is_bound(self):
        headers = {"User-Agent": "Smith", **self.headers}
        self.app.get("/", headers=headers)
        event_dict = self.logger_context()
        self.assertEqual(event_dict["path"], "/v0/")
        self.assertEqual(event_dict["method"], "GET")
        self.assertEqual(event_dict["code"], 200)
        self.assertEqual(event_dict["agent"], "Smith")
        self.assertIsNotNone(event_dict["uid"])
        self.assertIsNotNone(event_dict["time"])
        self.assertIsNotNone(event_dict["t"])
        self.assertEqual(event_dict["errno"], 0)
        self.assertNotIn("lang", event_dict)
        self.assertNotIn("headers", event_dict)
        self.assertNotIn("body", event_dict)

    def test_userid_is_none_when_anonymous(self):
        self.app.get("/")
        event_dict = self.logger_context()
        self.assertNotIn("uid", event_dict)

    def test_lang_is_not_none_when_provided(self):
        self.app.get("/", headers={"Accept-Language": "fr-FR"})
        event_dict = self.logger_context()
        self.assertEqual(event_dict["lang"], "fr-FR")

    def test_agent_is_not_none_when_provided(self):
        self.app.get("/", headers={"User-Agent": "webtest/x.y.z"})
        event_dict = self.logger_context()
        self.assertEqual(event_dict["agent"], "webtest/x.y.z")

    def test_errno_is_specified_on_error(self):
        self.app.get("/unknown", status=404)
        event_dict = self.logger_context()
        self.assertEqual(event_dict["errno"], 111)

    def test_basic_authn_type_is_bound(self):
        app = self.make_app({"multiauth.policies": "basicauth"})
        app.get("/mushrooms", headers={"Authorization": "Basic bWF0OjE="})
        event_dict = self.logger_context()
        self.assertEqual(event_dict["authn_type"], "basicauth")

    def test_request_id_is_taken_from_headers(self):
        self.app.get("/", headers={"X-Request-Id": "foo"})
        event_dict = self.logger_context()
        self.assertEqual(event_dict["rid"], "foo")

    def test_headers_and_body_when_level_is_debug(self):
        self.mocked.level = logging.DEBUG
        body = b'{"boom": 1}'
        self.app.post("/batch", body, headers=self.headers, status=400)
        event_dict = self.logger_context()
        self.assertEqual(
            event_dict["headers"],
            {
                "Authorization": "Basic bWF0OnNlY3JldA==",
                "Content-Length": "11",
                "Content-Type": "application/json",
                "Host": "localhost:80",
            },
        )
        self.assertEqual(event_dict["body"], body)

        self.maxDiff = None

        responseBody = event_dict["response"]["body"]
        self.assertEqual(json.loads(responseBody.decode("utf-8"))["error"], "Invalid parameters")
        responseHeaders = event_dict["response"]["headers"]
        self.assertEqual(
            sorted(responseHeaders.keys()),
            [
                "Access-Control-Expose-Headers",
                "Content-Length",
                "Content-Security-Policy",
                "Content-Type",
                "X-Content-Type-Options",
            ],
        )


class BatchSubrequestTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()

        patch = mock.patch("kinto.core.views.batch.subrequest_logger")
        self.subrequest_mocked = patch.start()
        self.addCleanup(patch.stop)

        patch = mock.patch("kinto.core.initialization.summary_logger")
        self.summary_mocked = patch.start()
        self.addCleanup(patch.stop)

        headers = {**self.headers, "User-Agent": "readinglist"}
        body = {
            "requests": [
                {"path": "/unknown", "headers": {"User-Agent": "foo"}},
                {"path": "/unknown2"},
            ]
        }
        self.app.post_json("/batch", body, headers=headers)

    def test_batch_global_request_is_preserved(self):
        args, kwargs = self.summary_mocked.info.call_args_list[-1]
        extra = kwargs["extra"]
        self.assertEqual(extra["code"], 200)
        self.assertEqual(extra["path"], "/v0/batch")
        self.assertEqual(extra["agent"], "readinglist")

    def test_batch_size_is_bound(self):
        args, kwargs = self.summary_mocked.info.call_args_list[-1]
        extra = kwargs["extra"]
        self.assertEqual(extra["batch_size"], 2)

    def test_subrequests_are_not_logged_as_request_summary(self):
        self.assertEqual(self.summary_mocked.info.call_count, 1)

    def test_subrequests_are_logged_as_subrequest_summary(self):
        self.assertEqual(self.subrequest_mocked.info.call_count, 2)
        args, kwargs = self.subrequest_mocked.info.call_args_list[-1]
        extra = kwargs["extra"]
        self.assertEqual(extra["path"], "/v0/unknown2")
        args, kwargs = self.subrequest_mocked.info.call_args_list[-2]
        extra = kwargs["extra"]
        self.assertEqual(extra["path"], "/v0/unknown")


class JsonFormatterTest(unittest.TestCase):
    def test_logger_name(self):
        JsonLogFormatter.init_from_settings({"project_name": "kintowe"})
        f = JsonLogFormatter()
        record = logging.LogRecord("app.log", logging.DEBUG, "", 0, "coucou", (), None)
        result = f.format(record)
        logged = json.loads(result)
        self.assertEqual(logged["Logger"], "kintowe")
        self.assertEqual(logged["Type"], "app.log")
        # See https://github.com/mozilla/mozilla-cloud-services-logger/issues/2
        self.assertEqual(logged["Fields"]["msg"], "coucou")



import re

from kinto.core.decorators import synchronized
from kinto.core.permission import PermissionBase


class Permission(PermissionBase):
    """Permission backend implementation in local process memory.

    Enable in configuration::

        kinto.permission_backend = kinto.core.permission.memory

    :noindex:
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.flush()

    def initialize_schema(self, dry_run=False):
        # Nothing to do.
        pass

    def flush(self):
        self._store = {}

    @synchronized
    def add_user_principal(self, user_id, principal):
        user_key = f"user:{user_id}"
        user_principals = self._store.get(user_key, set())
        user_principals.add(principal)
        self._store[user_key] = user_principals

    @synchronized
    def remove_user_principal(self, user_id, principal):
        user_key = f"user:{user_id}"
        user_principals = self._store.get(user_key, set())
        try:
            user_principals.remove(principal)
        except KeyError:
            pass
        if len(user_principals) == 0:
            if user_key in self._store:
                del self._store[user_key]
        else:
            self._store[user_key] = user_principals

    @synchronized
    def remove_principal(self, principal):
        for key, user_principals in self._store.items():
            if not key.startswith("user:"):
                continue
            try:
                user_principals.remove(principal)
            except KeyError:
                pass

    @synchronized
    def get_user_principals(self, user_id):
        # Fetch the groups the user is in.
        user_key = f"user:{user_id}"
        members = self._store.get(user_key, set())
        # Fetch the groups system.Authenticated is in.
        group_authenticated = self._store.get("user:system.Authenticated", set())
        return members | group_authenticated

    @synchronized
    def add_principal_to_ace(self, object_id, permission, principal):
        permission_key = f"permission:{object_id}:{permission}"
        object_permission_principals = self._store.get(permission_key, set())
        object_permission_principals.add(principal)
        self._store[permission_key] = object_permission_principals

    @synchronized
    def remove_principal_from_ace(self, object_id, permission, principal):
        permission_key = f"permission:{object_id}:{permission}"
        object_permission_principals = self._store.get(permission_key, set())
        try:
            object_permission_principals.remove(principal)
        except KeyError:
            pass
        if len(object_permission_principals) == 0:
            if permission_key in self._store:
                del self._store[permission_key]
        else:
            self._store[permission_key] = object_permission_principals

    @synchronized
    def get_object_permission_principals(self, object_id, permission):
        permission_key = f"permission:{object_id}:{permission}"
        members = self._store.get(permission_key, set())
        return members

    @synchronized
    def get_accessible_objects(self, principals, bound_permissions=None, with_children=True):
        principals = set(principals)
        candidates = []
        if bound_permissions is None:
            for key, value in self._store.items():
                _, object_id, permission = key.split(":", 2)
                candidates.append((object_id, permission, value))
        else:
            for pattern, perm in bound_permissions:
                id_match = ".*" if with_children else "[^/]+"
                regexp = re.compile(f"^{pattern.replace('*', id_match)}$")
                for key, value in self._store.items():
                    if key.endswith(perm):
                        object_id = key.split(":")[1]
                        if regexp.match(object_id):
                            candidates.append((object_id, perm, value))

        perms_by_object_id = {}
        for object_id, perm, value in candidates:
            if len(principals & value) > 0:
                perms_by_object_id.setdefault(object_id, set()).add(perm)
        return perms_by_object_id

    @synchronized
    def get_authorized_principals(self, bound_permissions):
        principals = set()
        for obj_id, perm in bound_permissions:
            principals |= self.get_object_permission_principals(obj_id, perm)
        return principals

    @synchronized
    def get_objects_permissions(self, objects_ids, permissions=None):
        result = []
        for object_id in objects_ids:
            if permissions is None:
                aces = [k for k in self._store.keys() if k.startswith(f"permission:{object_id}:")]
            else:
                aces = [f"permission:{object_id}:{permission}" for permission in permissions]
            perms = {}
            for ace in aces:
                # Should work with 'permission:/url/id:object:create'.
                permission = ace.split(":", 2)[2]
                perms[permission] = set(self._store[ace])
            result.append(perms)
        return result

    @synchronized
    def replace_object_permissions(self, object_id, permissions):
        for permission, principals in permissions.items():
            permission_key = f"permission:{object_id}:{permission}"
            if permission_key in self._store and len(principals) == 0:
                del self._store[permission_key]
            elif principals:
                self._store[permission_key] = set(principals)
        return permissions

    @synchronized
    def delete_object_permissions(self, *object_id_list):
        to_delete = []
        for key in self._store.keys():
            object_id = key.split(":")[1]
            for pattern in object_id_list:
                regexp = re.compile(f"^{pattern.replace('*', '.*')}$")
                if regexp.match(object_id):
                    to_delete.append(key)
        for k in to_delete:
            del self._store[k]


def load_from_config(config):
    return Permission()



from kinto.core.testing import unittest

from .support import BaseWebTest


BUCKET_URL = "/buckets/blog"
COLLECTION_URL = "/buckets/blog/collections/articles"


SCHEMA = {
    "title": "Collection schema",
    "type": "object",
    "required": ["displayFields"],
    "properties": {
        "uiSchema": {"type": "object"},
        "displayFields": {"type": "array", "items": {"type": "string"}},
    },
}

VALID_COLLECTION = {"uiSchema": {"schema": {"ui:widget": "hidden"}}, "displayFields": ["name"]}


class DeactivatedSchemaTest(BaseWebTest, unittest.TestCase):
    def test_schema_should_be_json_schema(self):
        newschema = {**SCHEMA, "type": "Washmachine"}
        resp = self.app.put_json(
            BUCKET_URL,
            {"data": {"collection:schema": newschema}},
            headers=self.headers,
            status=400,
        )
        error_msg = "'Washmachine' is not valid under any of the given schemas"
        self.assertIn(error_msg, resp.json["message"])

    def test_records_are_not_invalid_if_do_not_match_schema(self):
        self.app.put_json(
            BUCKET_URL, {"data": {"collection:schema": SCHEMA}}, headers=self.headers
        )
        self.app.put_json(
            COLLECTION_URL, {"data": {"displayFields": 42}}, headers=self.headers, status=201
        )


class BaseWebTestWithSchema(BaseWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["experimental_collection_schema_validation"] = "True"
        return settings


class CollectionValidationTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        resp = self.app.put_json(
            BUCKET_URL, {"data": {"collection:schema": SCHEMA}}, headers=self.headers
        )
        self.collection = resp.json["data"]

    def test_empty_metadata_can_be_validated(self):
        self.app.post_json(BUCKET_URL + "/collections", headers=self.headers, status=400)
        self.app.post_json(
            BUCKET_URL + "/collections", {"data": {}}, headers=self.headers, status=400
        )

    def test_collections_are_valid_if_match_schema(self):
        self.app.put_json(
            COLLECTION_URL, {"data": VALID_COLLECTION}, headers=self.headers, status=201
        )

    def test_collections_are_invalid_if_do_not_match_schema(self):
        self.app.put_json(
            COLLECTION_URL, {"data": {"displayFields": 42}}, headers=self.headers, status=400
        )


SCHEMA_UNRESOLVABLE = {"properties": {"displayFields": {"$ref": "#/definitions/displayFields"}}}


class CollectionUnresolvableTest(BaseWebTestWithSchema, unittest.TestCase):
    def setUp(self):
        super().setUp()
        resp = self.app.put_json(
            BUCKET_URL, {"data": {"collection:schema": SCHEMA_UNRESOLVABLE}}, headers=self.headers
        )
        self.collection = resp.json["data"]

    def test_unresolvable_errors_handled(self):
        self.app.put_json(
            COLLECTION_URL, {"data": {"displayFields": 42}}, headers=self.headers, status=400
        )
